/*
Main startup file for Ian Wright's minder system.
Modified to use newkit: 7 Sep 2008

Altered to use 'define :rulesystem' syntax instead of using
list sim_rulesets, as in old version of SimAgent

*/


-367169140 -> ranseed;
ranseed => vars oldr; ranseed -> oldr;
max(popmemlim, 1000000) -> popmemlim;
constant ianpopdir = sys_fname_path(popfilename);
extend_searchlist(ianpopdir, popuseslist) -> popuseslist;

;;; uses profile;

uses newkit
uses sim_agent;
uses sim_geom;
uses polys;

true -> popgctrace;
global vars old_agents, all_agents;

;;; rulesets----------------------------------------------------

global vars
    trsim_perception_rules = [],
    trsim_belief_rules = [],
    trsim_actor_action_rules = [],
    trsim_bug_action_rules = [],
    trsim_bug_rules = [],
    trsim_filter_motives = [],
    trsim_get_active_motives = [],
    trsim_order_motives = [],
    trsim_runplan_rules = [],
    trsim_ditch_rules = [],
;

global vars ;;; Sloman and Beaudoin's generactivators
    G_low_charge = [],
    G_bug_dead = [],
    G_is_frightened = [],
    G_patrol_ditches = [],
    G_near_ditch = [],
    G_no_maginot = [],
    G_below_maginot = [],
    G_default = [],
    G_condense_motives = []     ;;; `cognitive reflex'
;

define :rulesystem trsim_actor_rulesets;
    include:    trsim_perception_rules
    include:    trsim_belief_rules
    ;;; suite of asynchronous, parallel generactivators
    include:        G_low_charge
    include:        G_bug_dead
    ;;;;include:        G_is_frightened
    include:        G_patrol_ditches
    include:        G_near_ditch
    include:        G_no_maginot
    include:        G_below_maginot
    include:        G_default
    ;;;;include:        G_condense_motives
    include:    trsim_filter_motives
    include:    trsim_get_active_motives
    include:    trsim_order_motives
    include:    trsim_runplan_rules
    include:    trsim_actor_action_rules
enddefine;

define :rulesystem trsim_bug_rulesets;
    include:    trsim_perception_rules
    include:    trsim_bug_rules
    include:    trsim_bug_action_rules
enddefine;

define :rulesystem trsim_ditch_rulesets;
    include:    trsim_perception_rules
    include:    trsim_ditch_rules
enddefine;

;;; classes-----------------------------------------------------

constant world_size = 100;

define :class trsim_agent; is sim_agent;
    slot trsim_heading = 90;
    slot trsim_speed == 1;
    slot trsim_location = conspair(random(world_size - 20) + 10,
        random(world_size - 20) + 10);
    slot trsim_status == "alive";
    slot trsim_held = false;
    slot trsim_type;
    slot trsim_id;
    slot trsim_charge = 100;
    slot trsim_clock = 0;
    slot sim_rulesystem = [];
    slot sim_sensors = [{sim_sense_agent 20}];
    slot trsim_clock = 0;

    slot trsim_xor = []; ;;; will contain a list of old values of
                         ;;; slots that have significance for graphic
                         ;;; updating.
enddefine;

define :mixin trsim_bar;
    slot bar_length = 20;
    slot bar_width = 0;
    slot bar_current_poly;
enddefine;

define :class trsim_bar_agent; is trsim_agent trsim_bar;
    slot bar_end1a_location;
    slot bar_end1b_location;
    slot bar_end2a_location;
    slot bar_end2b_location;
    slot trsim_type == "bar";
    slot trsim_heading = random(360);
    slot sim_rulesystem = [];
    slot sim_sensors = [];
enddefine;

define :class trsim_wall_agent; is trsim_bar_agent;
    slot trsim_type == "bar"; ;;; treat walls like bars
enddefine;

define :class trsim_ditch_agent; is trsim_agent trsim_bar;
    slot trsim_heading = 90;
    slot trsim_type = "ditch";
    slot sim_rulesystem = trsim_ditch_rulesets;
    slot sim_sensors = [{sim_sense_agent ^world_size}];
enddefine;

define :class trsim_static_agent; is trsim_agent;
    slot trsim_type = "static";
    slot sim_rulesystem = [];
    slot sim_sensors = [];
enddefine;

define :class trsim_actor_agent; is trsim_agent;
    slot trsim_location = conspair(50,50);
    slot trsim_max_speed = 5;
    slot trsim_type = "actor";
    slot trsim_active_motive = [];
    slot sim_rulesystem = trsim_actor_rulesets;
    slot sim_cycle_limit = 10;
enddefine;

define :class trsim_bug_agent; is trsim_agent;
    slot trsim_type = "bug";
    slot sim_rulesystem = trsim_bug_rulesets;
enddefine;

;;;constant window_size = 700;
constant window_size = 350;
constant boundary = [ [0 0] [0 ^world_size] [^world_size ^world_size]
    [^world_size 0] ];

;;; instances ---------------------------------------------------

define :instance actor1:trsim_actor_agent;
    trsim_id = 'a';
    trsim_speed = 1;
enddefine;

define :instance actor2:trsim_actor_agent;
    trsim_id = 'b';
    trsim_speed = 1;
    trsim_location = conspair(90, 10);
enddefine;

define :instance wall1:trsim_wall_agent;
    bar_length = world_size - 1;
    bar_width = 1;
    trsim_heading = 90;
    trsim_location = conspair(0, world_size div 2);
    trsim_id = 'wall';
enddefine;

define :instance wall2:trsim_wall_agent;
    bar_length = world_size;
    bar_width = 1;
    trsim_heading = 0;
    trsim_location = conspair(world_size div 2, 0);
    trsim_id = 'wall';
enddefine;

define :instance wall3:trsim_wall_agent;
    bar_length = world_size - 1;
    bar_width = 1;
    trsim_heading = 90;
    trsim_location = conspair(world_size, world_size div 2);
    trsim_id = 'wall';
enddefine;

define :instance wall4:trsim_wall_agent;
    bar_length = world_size;
    bar_width = 1;
    trsim_heading = 0;
    trsim_location = conspair(world_size div 2, world_size);
    trsim_id = 'wall';
enddefine;

define :instance ditch1:trsim_ditch_agent;
    bar_length = world_size div 2;
    bar_width = 4;
    trsim_location = conspair(5, 30);
    trsim_id = 'ditch';
    trsim_held = true;
enddefine;

define :instance ditch2:trsim_ditch_agent;
    bar_length = world_size div 2;
    bar_width = 4;
    trsim_location = conspair(world_size - 5, 30);
    trsim_id = 'ditch';
    trsim_held = true;
enddefine;

define :instance gas1:trsim_static_agent;
    trsim_id = 'F';
    trsim_location = conspair(world_size div 2, 10);
enddefine;

define :instance exit1:trsim_static_agent;
    trsim_id = 'E';
    trsim_location = conspair(world_size div 2, world_size - 2);
enddefine;

[
    actor1
;;;    actor2
    wall1 wall2 wall3 wall4
    ditch1 ditch2
    gas1 exit1
] -> all_agents;

uses gi;

;;; general utilities ------------------------------------------

define :method sim_sense_agent(a1:sim_object, a2:sim_object, dist);
    ;;; Default sensor for detecting other agents
    ;;; return information about name of other agent and distance
    ;;; Users will probably want to redefine this. E.g. it might
    ;;; include a test for direction - using cone of visibility.
    ;;; If a2 is not to be detected, then return nothing
    [new_sense_data
        time ^(a1.trsim_clock)
        name ^(a2.sim_name)
        type ^(a2.trsim_type)
        status ^(a2.trsim_status)
        distance ^dist
        x ^(front(a2.trsim_location))
        y ^(back(a2.trsim_location))
        id ^(a2.trsim_id)
        charge ^(a2.trsim_charge)
        held % if a2.trsim_held then "true" else "false" endif; %
    ]
enddefine;

define :method sim_sense_agent(a1:sim_object, a2:trsim_bar_agent, dist);
    unless istrsim_bug_agent(a1) or istrsim_ditch_agent(a1) then
        [new_sense_data
            time ^(a1.trsim_clock)
            name ^(a2.sim_name)
            type ^(a2.trsim_type)
            status ^(a2.trsim_status)
            distance ^dist
            x ^(front(a2.trsim_location))
            y ^(back(a2.trsim_location))
            id ^(a2.trsim_id)
            held % if a2.trsim_held then "true" else "false" endif; %

            poly_space ^(a2.bar_current_poly)
            heading ^(a2.trsim_heading)
        ]
    endunless;
enddefine;

define :method sim_sense_agent(a1:sim_object, a2:trsim_ditch_agent, dist);
    unless istrsim_bug_agent(a1) or istrsim_ditch_agent(a1) then
        [new_sense_data
            time ^(a1.trsim_clock)
            name ^(a2.sim_name)
            type ^(a2.trsim_type)
            status ^(a2.trsim_status)
            distance ^dist
            x ^(front(a2.trsim_location))
            y ^(back(a2.trsim_location))
            id ^(a2.trsim_id)
            held % if a2.trsim_held then "true" else "false" endif; %

            poly_space ^(a2.bar_current_poly)
            heading ^(a2.trsim_heading)
        ]
    endunless;
enddefine;

define procedure get_the_distance(loc1, loc2) -> dist;
lvars loc1, loc2, dist;

    sim_distance_from(destpair(loc1), destpair(loc2)) -> dist;

enddefine;

define :method sim_distance(a1:sim_object, a2:sim_object) -> dist;
    ;;; Compute distance between two agents.
    lvars a1, a2, dist;
    get_the_distance(a1.trsim_location, a2.trsim_location) -> dist;
enddefine;

define :method sim_distance(a1:sim_object, a2:trsim_bar_agent) -> dist;
lvars a1, a2, dist, a1loc = a1.trsim_location;

    min(
        min(get_the_distance(a1loc, a2.trsim_location),
            get_the_distance(a1loc, a2.bar_end1a_location)),
                get_the_distance(a1loc, a2.bar_end2a_location)) -> dist

enddefine;

;;; Reduce tracing

define vars procedure sim_scheduler_pausing_trace(objects, cycle);
	;;; user definable
	lvars objects, cycle;
	if cycle mod 20 == 0 then
		pr('\n end of cycle ==' >< cycle >< '==\n');
		readline() ->;
	endif;
enddefine;

;;; Here are non-printing versions of trace functions.
define :method sim_agent_messages_out_trace(agent:sim_agent);
    lvars agent;
/*
    lvars messages = sim_out_messages(agent);
    [New messages ^(sim_name(agent)) ^messages] ==>
*/
enddefine;

define :method sim_agent_messages_in_trace(agent:sim_agent);
    lvars agent;
/*
    lvars messages = sim_out_messages(agent);
    [New messages ^(sim_name(agent)) ^messages] ==>
*/
enddefine;

define :method sim_agent_actions_out_trace(agent:sim_object);
    lvars agent;
/*
    lvars actions = sim_actions(agent);
    [New actions ^(sim_name(agent)) ^actions] ==>
*/
enddefine;

define :method sim_agent_rulefamily_trace(object:sim_object, rulefamily);
    lvars object, rulefamily;
/*
    ['Try rulefamily' ^rulefamily 'with object' ^(sim_name(object))]==>
    pr('With Data:\n'); prb_print_table( (sim_data(object)) );
*/
enddefine;

define :method sim_agent_endrun_trace(agent:sim_object);
/*
    unless islist(agent.sim_rulesystem) then
        ['Data in' ^(sim_name(agent)):]==>
        prb_print_table(sim_data(agent));
    endunless;
*/
enddefine;

define :method sim_agent_terminated_trace(object:sim_object,
            stopped, max_cycles, used);
enddefine;

;;; Specialise the print method
define :method print_instance(item:trsim_agent);
    printf( '<agent %P >', [% sim_name(item) %])
enddefine;

;;; trsim_perception_rules----------------------------------------

;;; rules---------------------------------------------------------

define :ruleset trsim_perception_rules;

rule see_first
     ==>
    [VARS c_time]
    [POP11 sim_myself.trsim_clock - 1 -> c_time;]
    [NOT new_sense_data time ?c_time ==]

rule see_last
     ==>
    [STOP]
enddefine;

;;; other rules -------------------------------------------------

uses actor_rules;
uses bug_rules;
uses ditch_rules;

;;; setups -------------------------------------------------------

define make_bugs(bug_num) -> bug_list;
lvars bug_num, bug_list, temp, nm = 1;

    1 -> gensym("bug");
    [%
        repeat bug_num times
            gensym("bug") -> temp;

            ;;; declare it as a variable
            ident_declare(temp, 0, false);

            instance trsim_bug_agent;
                sim_name = temp;
                trsim_id = nm >< '';
                trsim_speed = 1;
            endinstance ->> valof(temp);

            nm + 1 -> nm;
        endrepeat;
    %] -> bug_list;

enddefine;

define make_bars(bar_num) -> bar_list;
lvars bar_num, bar_list, temp;

    1 -> gensym("bar");
    [%
        repeat bar_num times
            gensym("bar") -> temp;
            ;;; declare it as a variable
            ident_declare(temp, 0, false);

            instance trsim_bar_agent;
                sim_name = temp;
                trsim_id = 'bar';   ;;; why a string?
            endinstance ->> valof(temp);
        endrepeat;
    %] -> bar_list;

enddefine;

define procedure make_objects;
lvars item, bug_list, bar_list;
;;;lconstant num_bugs = 6;
;;;lconstant num_bars = 20;
lconstant num_bugs = 3;
lconstant num_bars = 10;

    [%for item in all_agents do
            ;;; declare it as a variable
            ident_declare(item, 0, false);
        item-> sim_name(valof(item)); valof(item) endfor
    %] -> all_agents;
    make_bugs(num_bugs) -> bug_list;
    make_bars(num_bars) -> bar_list;
    all_agents <> (bug_list <> bar_list) -> all_agents;

enddefine;

define procedure setup;
lvars ag, x1, y1, item;

    true -> prb_chatty;
    false -> prb_chatty;

    true -> prb_walk;
    false -> prb_walk;

    true -> prb_show_conditions;
    false -> prb_show_conditions;
/*
[
trsim_do_move_closure_a
trsim_do_move_closure_holding
trsim_do_rotate_closure_a
trsim_do_setspeed
trsim_do_grab_object_closure
trsim_do_drop_closure
trsim_do_rotate_bar_closure
trsim_do_charge_closure
trsim_do_dismiss_closure
act_last_a
] -> prb_show_conditions;
[
TR_search_r1
TR_search_r2
TR_search_r3
TR_search_r4
TR_search_last
] -> prb_show_conditions;
[
TR_open_door_r1
TR_open_door_r2
TR_open_door_r3
TR_open_door_r4
TR_open_door_last
] -> prb_show_conditions;
[
TR_dismiss_object_r1
TR_dismiss_object_r2
TR_dismiss_object_r3
TR_dismiss_object_last
] -> prb_show_conditions;
[
default_behaviour
motive_done
trp_done
make_trp
runTRP
last_actor_rules
] -> prb_show_conditions;
[
TR_amble_r1
TR_amble_r2
TR_amble_r3
TR_amble_r4
TR_amble_last
] -> prb_show_conditions;
[grule_low_charge_remove_r2] -> prb_show_conditions;
[
TR_put_safe_r1
TR_put_safe_r2
TR_put_safe_r3
TR_put_safe_r4
TR_put_safe_r5
TR_put_safe_r6
TR_put_safe_r7
TR_put_safe_last
] -> prb_show_conditions;
[
TR_drop_r3
] -> prb_show_conditions;
[
become_normal
become_thug
calm_down
attack_bug
bug_rules_last
] -> prb_show_conditions;
[
grule_low_charge_add
] -> prb_show_conditions;
[
init_descriptor_table
create_descriptor_table
condense_motives_last
] -> prb_show_conditions;
[
rule decide_low_charge_low
rule decide_low_charge_ok
rule decide_low_charge_high
rule init_cost
rule SP_decide_last
] -> prb_show_conditions;
*/

    true -> prb_allrules;
    ;;;false -> prb_allrules;

    true -> prb_repeating;
    false -> prb_repeating;

    make_objects();
    all_agents =>
    create_window();
    rc_set_background('White');
    rc_set_ink_color('Black');
    for ag in all_agents do
        draw_move_agent(ag)
    endfor;

enddefine;

define procedure go(cycle);
lvars agn, ago;

	false -> popgctrace;
    sys_unlock_heap();
    sysgarbage();
    kill();
    setup();
    sysgarbage();
    sys_lock_heap();

    sim_scheduler(all_agents, cycle);

enddefine;

pr('type\ngo(1000);');
