;;; TO DO:

;;; When there is more than one active motive at a time,
;;; will have to change the first condition of TR rules,
;;; otherwise they may unify with the incorrect motive,
;;; i.e. another active motive.

global vars ;;; Nilsson's Teleo-Reactive Programs
    TRP_goto = [],          ;;; goto loc
    TRP_amble = [],         ;;; goto loc while avoiding obstacles
    TRP_goto_object = [],   ;;; goto object (search if no information)
    TRP_grab_object = [],   ;;; pickup object
    TRP_take_object = [],   ;;; takes an object to another object
    TRP_drop = [],          ;;; drop object
    TRP_search = [],        ;;; search domain
    TRP_place_bar = [],     ;;; drop bar at loc with correct heading
    TRP_grab_wall = [],     ;;; pickup a bar to use as a wall
    TRP_make_wall = [],     ;;; choose a bar and make a wall
    TRP_shut_door = [],     ;;; shut the room door
    TRP_open_door = [],     ;;; open the room door
    TRP_charge_object = [], ;;; recharge an object at charge point
    TRP_dismiss_object = [],;;; dismiss an object at exit point
    TRP_put_safe = []       ;;; places an object at a safe distance from
                            ;;; another object
;

global vars ;;; special plan steps
    SP_leave = [],          ;;; used when motive must not execute more
                            ;;; plan steps, but shouldn't be removed
    SP_get_plan = [],       ;;; plans for a motive
    SP_condense_motives = [],
    SP_decide = []
;

global vars TR_index =
[
    goto            TRP_goto
    amble           TRP_amble
    goto_object     TRP_goto_object
    grab_object     TRP_grab_object
    take_object     TRP_take_object
    drop            TRP_drop
    search          TRP_search
    place_bar       TRP_place_bar
    grab_wall       TRP_grab_wall
    make_wall       TRP_make_wall
    shut_door       TRP_shut_door
    open_door       TRP_open_door
    charge_object   TRP_charge_object
    dismiss_object  TRP_dismiss_object
    put_safe        TRP_put_safe

    leave           SP_leave
    get_plan        SP_get_plan
    condense        SP_condense_motives
    decide          SP_decide
];

global constant bounce_ratio = 10;
global constant ap_room_loc = conspair(50,75);
global constant ap_offset = 9; ;;; half length of domain bars
global constant RL_max_motives = 3;

;;; trsim_belief_rules-----------------------------------------------

;;; utilities ----------------------------------------------------------

define procedure heading_close_enough(h1, h2) -> bool;
lvars h1, h2, bool;
lconstant closeness = 1;

    (abs(h1 - h2) <= closeness)
    or
    (abs(max(h1, h2) - 360 - min(h1, h2)) <= closeness)  -> bool

enddefine;

/*
    heading_close_enough(100, 102) =>
    heading_close_enough(360, 2) =>
    heading_close_enough(2, 359) =>
*/

define procedure is_right_orientation(h1, h2) -> bool;
lvars h1, h2, bool;

    heading_close_enough(h1, h2) or
            heading_close_enough(h1, rotate_angle(h2, 180)) -> bool

enddefine;

/*
    is_right_orientation(280, 91) =>

*/

define procedure difference_in_heading(h1at, h2to) -> df;
lvars h1at, h2to, df, df1;

    h2to - h1at -> df1;
    rotate_angle(h2to, 180) -> h2to;
    h2to - h1at -> df;
    abs(df1) -> df1;
    abs(df) -> df;
    if df1 < df then df1 -> df endif;

enddefine;

/*
    difference_in_heading(90, 45) =>
    difference_in_heading(200, 0) =>
    difference_in_heading(90, 92) =>
    difference_in_heading(90, 89) =>
    difference_in_heading(90, 270) =>
    difference_in_heading(200, 360) =>
*/

;;; rules------------------------------------------------------------

/*
    Types of belief:

    Sensory-based
    [belief time t name n ... ]
        - belief about a basic object in environment,
          such as bar, bug, wall, ditch

    Agent-based (based on sensory-based beliefs)
    [belief name %oneof([north east south west])% wall is obj]
        - belief about a bar that is part of an agent constructed
          room.

    Temporary agent-based
    [belief name %oneof([[candidate_wall is] [candidate_walls are]])% obj]
        - interim beliefs for reasoning about room construction

    A belief of the form [belief == [m_conditions ??conds]] is a
    belief that is automatically maintained, i.e. conds
    contains the conditions that must be false in order for
    the belief to remain true. If conds ever evaluates to
    true then the belief is removed. Therefore, when a new
    belief is created, and it is to be maintained, the
    programmer must specify these conditions.

    NOTE: use NOT, not DEL to remove beliefs, as m_conditions
        unify in condition parts of rules.
*/

;;; lvars Id, Xb, Yb;
define proc_m(Xb, Yb, Id);
    (Id /= 'wall') and (Id /= 'ditch')
    and
        sim_distance_from(destpair(sim_myself.trsim_location), Xb,Yb)
          ;;;  < last(last(sim_myself.sim_sensors))
            < 5 ;;; i.e., if can't see object when 5 steps from
                ;;; last known location, then agent doesn't know
                ;;; where it is.

;;; can put this code in WHERE

enddefine;

define :ruleset trsim_belief_rules;

rule belief_maintenance_r1
;;; generic belief maintenance rule -- maintains beliefs
;;; that have their maintenance conditions stored with them
    [belief time = name ?Obj_name == [m_conditions ?Mconds]]
    [ALL ?Mconds]
     ==>
    ;;; don't necessarily want to remove belief; rather,
    ;;; change the information (but don't delete all the
    ;;; information.

    [NOT belief time = name ?Obj_name ==]
    [SAY [Removed belief about ?Obj_name]]

rule update_beliefs
    ;;; takes new sensory input and updates beliefs
    [new_sense_data time ?Time name ?Obj_name ??Info]
    [belief time ?Time2 name ?Obj_name == [m_conditions ??Mconds]]
    [WHERE Time2 < Time ]
     ==>
    [NOT belief time ?Time2 name ?Obj_name ==]
    [belief time ?Time name ?Obj_name ??Info [m_conditions ??Mconds]]

rule create_new_belief
    ;;; makes a new belief if sensory input contains
    ;;; Info about a newly encountered object
    [new_sense_data time ?Time name ?Obj_name ??Info]
    [NOT belief time = name ?Obj_name ==]
     ==>
    [belief time ?Time name ?Obj_name ??Info
        [m_conditions [
            [belief == name ?Obj_name == x ?Xb y ?Yb id ?Id ==]
            [WHERE proc_m(Xb, Yb, Id)]
            [NOT new_sense_data == name ?Obj_name ==] ] ] ]

rule belief_rules_last
    ;;; no more sensory data to process
     ==>
    ;;; [POP11 (sim_myself.sim_data)("belief") ==> ]
    [STOP]
enddefine;

;;; G_low_charge----------------------------------------------------

global constant chg_threshold = 80;

;;; utilities-------------------------------------------------------

define procedure low_charge_insistence(chg_level) -> ins;
lvars chg_level, ins;
/*
    -1/200 x + 0.5
*/
    -0.005 * chg_level + 0.5 -> ins;
    if ins < 0 then 0 -> ins endif;
    if ins > 1 then 1 -> ins endif;
enddefine;

/*
    low_charge_insistence(90) =>

*/

;;; rules-----------------------------------------------------------

;;; At this point, agent's database will contain updated beliefs.

;;; ADD
define :ruleset G_low_charge;

rule grule_low_charge_remove_r1
    [MOTIVE motive [low_charge ?Obj_name] ==][ ->> Motive ]
    [not_exists ?Obj_name]
     ==>
    [DEL ?Motive]

rule grule_low_charge_remove_r2
    [MOTIVE motive [low_charge ?Obj_name] ==][ ->> Motive ]
    [belief == name ?Obj_name == status dead ==]
    ==>
    [DEL ?Motive]

rule grule_low_charge_remove_r3
    [MOTIVE motive == status active motive [low_charge ?Obj_name]
        insistence ?Insistence ==] [ ->> Motive ]
    [WHERE Insistence = 0 ]
==>
    [DEL ?Motive]

rule grule_low_charge_add
    [belief == name ?Obj_name == status alive == charge ?Charge ==]
    [WHERE Charge < chg_threshold ]
    [NOT MOTIVE motive [low_charge ?Obj_name] ==]
     ==>
    [LVARS insist]
    [POP11 low_charge_insistence(Charge) -> insist; ]
    [MOTIVE motive [low_charge ?Obj_name] insistence ?insist status sub]

;;; CHANGE
rule grule_low_charge_change
    [MOTIVE motive [low_charge ?Obj_name] insistence ?Insistence ==]
    [belief == name ?Obj_name == charge ?Charge ==]
    [LVARS insist]
    [WHERE low_charge_insistence(Charge) -> insist;
           Charge /= insist;
    ]
     ==>
    [MODIFY 1 insistence ?insist]

rule grules_last
     ==>
    [STOP]
enddefine;

;;; G_below_maginot----------------------------------------------

global vars G_below_maginot = [];
global constant below_maginot_ins = 0.2;

;;; utilities-----------------------------------------------------

define procedure below_maginot_insistence() -> ins;
lvars ins;
    below_maginot_ins -> ins
enddefine;

;;; rules--------------------------------------------------------

define :ruleset G_below_maginot;

rule grule_below_maginot_add
    [belief == name ?Obj type bug status alive == x ?X y ?Y ==]
    [belief name first wall is ==]
    [belief name second wall is ==]
    [belief name third wall is ==]
    [belief name fourth wall is ?Wall [at ?X1 ?Y1 ==] ==]
    [WHERE Y < Y1 ] ;;; i.e., bug is below maginot line
    [NOT MOTIVE motive [below_maginot ?Obj] ==]
==>
    [LVARS insist]
    [POP11 below_maginot_insistence() -> insist; ]
    [MOTIVE motive [below_maginot ?Obj] insistence ?insist status sub]

enddefine;

;;; G_bug_dead -------------------------------------------------

global constant bug_dead_ins = 0.15;

;;; utilities --------------------------------------------------

define procedure bug_dead_insistence() -> ins;
lvars ins;
    bug_dead_ins -> ins;
enddefine;

;;; rules ------------------------------------------------------

define :ruleset G_bug_dead;

/*
    ;;; not required as rationale will remain
    ;;; unchanged until dismissal
define :rule grule_bug_dead_remove in G_bug_dead
enddefine;
*/

rule grule_bug_dead_add
    [belief == name ?Obj_name type bug == status dead ==]
    [NOT MOTIVE motive [bug_dead ?Obj_name] ==]
     ==>
    [LVARS insist]
    [POP11 bug_dead_insistence() -> insist; ]
    [MOTIVE motive [bug_dead ?Obj_name] insistence ?insist status sub]

/*
    ;;; not required as insistence does not change
define :rule grule_bug_dead_change in G_bug_dead
enddefine;
*/

rule grules_last
     ==>
    [STOP]
enddefine;

;;; G_is_frightened ---------------------------------------------------------

global constant frightened_ins = 0.2;

;;; utilities-------------------------------------------------------

define procedure frightened_insistence() -> ins;
lvars ins;
    frightened_ins -> ins
enddefine;

;;; rules-----------------------------------------------------------

define :ruleset G_is_frightened;

rule grule_frightened_remove
    [MOTIVE motive [is_frightened ?Obj_name] ==][ ->> Motive]
    [belief == name ?Obj_name type bug == status ?Status ==]
    [WHERE Status /= "frightened"]
        ==>
    [DEL ?Motive]

rule grule_frightened_add
    [belief == name ?Obj_name type bug == status frightened ==]
    [NOT MOTIVE motive [is_frightened ?Obj_name] ==]
        ==>
    [LVARS insist]
    [POP11 frightened_insistence() -> insist; ]
    [MOTIVE motive [is_frightened ?Obj_name] insistence ?insist status sub]

rule grules_last
     ==>
    [STOP]
enddefine;

;;; G_patrol_ditches------------------------------------------------

global constant patrol_ditch_time = 200; ;;; cycles
global constant patrol_ditch_max_ins = 0.15;

;;; utilities-------------------------------------------------------

define procedure patrol_insistence(thetime) -> ins;
lvars ins, thetime;
    (patrol_ditch_max_ins / patrol_ditch_time) * thetime -> ins;
    if ins > patrol_ditch_max_ins then patrol_ditch_max_ins -> ins endif;
    if ins < 0 then 0 -> ins endif;
enddefine;

/*
    patrol_insistence(20) =>

*/

;;; rules-----------------------------------------------------------

;;; If actor has been near a particular ditch for a specified amount
;;; of time, then generate a motive to visit it.

define :ruleset G_patrol_ditches;

rule grule_patrol_ditches_add
    [belief time ?Time name ?Obj_name type ditch ==]
    [NOT MOTIVE motive [visit ?Obj_name] ==]
    [LVARS thetime]
    [WHERE
        sim_myself.trsim_clock - Time -> thetime;
        thetime > patrol_ditch_time
    ]
        ;;; may need to altered by learning
==>
    [LVARS insist]
    [POP11 patrol_insistence(thetime) -> insist; ]
    [MOTIVE motive [visit ?Obj_name] insistence ?insist status sub]

rule grules_last
     ==>
    [STOP]
enddefine;

;;; G_near_ditch----------------------------------------------------

;;; generates a motive if bug is near a ditch

global constant near_ditch_proximity = 15;
    ;;; may need to modify this through learning, e.g.
    ;;; if bugs suddenly get faster

;;; utilities-------------------------------------------------------

define procedure is_near_ditch(ditch_desc, px, py, x, y, dist) -> bool;
lvars ditch_desc, x, y, px, py, bool = false, point, dist;
lvars plist;

    [^px ^py] :: ditch_desc.poly_points -> plist;
    for point in plist do
        if get_the_distance(conspair(point(1),point(2)), conspair(x,y)) < dist then
            true -> bool;
            quitloop;
        endif;
    endfor;

enddefine;

define procedure near_ditch_insistence(ditch_desc, px, py, x, y) -> ins;
lvars ditch_desc, x, y, ins, point, min_dist = world_size, dist;
lvars plist;

    [^px ^py] :: ditch_desc.poly_points -> plist;
    for point in plist do
        get_the_distance(conspair(point(1),point(2)), conspair(x,y)) -> dist;
        if dist < min_dist then dist -> min_dist endif;
    endfor;
    ;;; want insistence to be in the range 0.2 -> 1
    ;;; (may need to compare with low_charge insistence which ranges
    ;;; from 0 -> 0.5)
    (-0.8 / near_ditch_proximity) * min_dist + 1 -> ins;
    if ins > 1 then 1 -> ins endif;
    if ins < 0 then 0 -> ins endif;
enddefine;

/*
    -0.04 * 2 + 1 =>
*/

;;; rules-----------------------------------------------------------

define :ruleset G_near_ditch;
;;; note use poly_points of ditch to determine proximity

rule grule_near_ditch_remove_r1
;;; remove motive if the bug has died
    [MOTIVE motive [near_ditch ?Obj2 ?Obj1] ==][ ->> Motive ]
    [belief == name ?Obj1 type bug status dead ==]
==>
    [DEL ?Motive]

rule grule_near_ditch_remove_r2
    [MOTIVE motive [near_ditch ?Obj2 ?Obj1] insistence ?Insistence ==]
        [ ->> Motive ]
    [WHERE Insistence = 0 ]
==>
    [DEL ?Motive]
    ;;; need this rule because motive may not complete due to
    ;;; interruption; however, when motive is tried again the
    ;;; bug has moved and is no longer in danger zone, i.e.
    ;;; insistence is zero. This removes motive, otherwise it
    ;;; will linger

vars Polygon;
rule grule_near_ditch_add
    [belief == name ?Obj1 type bug status alive == x ?X y ?Y ==
        held false ==]
    [belief == name ?Obj2 type ditch == x ?X1 y ?Y1 == poly_space ?Polygon ==]
    [WHERE
        is_near_ditch(Polygon, X1, Y1, X, Y, near_ditch_proximity) ]
    [NOT MOTIVE motive [near_ditch ?Obj2 ?Obj1] ==]
==>
    [LVARS insist]
    [POP11 near_ditch_insistence(Polygon, X1, Y1, X, Y) -> insist; ]
    [MOTIVE motive [near_ditch ?Obj2 ?Obj1] insistence ?insist status sub]

rule grule_near_ditch_change
    [MOTIVE motive [near_ditch ?Obj2 ?Obj1] insistence ?Insistence ==]
    [belief == name ?Obj1 type bug status alive == x ?X y ?Y ==
        held false ==]
    [belief == name ?Obj2 type ditch == x ?X1 y ?Y1 == poly_space ?Polygon ==]
    [LVARS insist]
    [WHERE
        near_ditch_insistence(Polygon, X1, Y1, X, Y) -> insist;
        insist /= Insistence ]
==>
    [MODIFY 1 insistence ?insist]

rule grules_last
     ==>
    [STOP]
enddefine;

;;; G_no_maginot--------------------------------------------------------

global constant no_maginot_ins = 0.05;

;;; utilities-----------------------------------------------------------

define procedure no_maginot_insistence() -> ins;
lvars ins;
    no_maginot_ins -> ins
enddefine;

;;; rules---------------------------------------------------------------

define :ruleset G_no_maginot;

rule grule_no_maginot_add
    [NOT MOTIVE motive [no_maginot] ==]
    [NOT belief name first wall is ==]
    [NOT belief name second wall is ==]
    [NOT belief name third wall is ==]
    [NOT belief name fourth wall is ==]
==>
    [LVARS insist]
    [POP11 no_maginot_insistence() -> insist]
    [MOTIVE motive [no_maginot] insistence ?insist status sub]

rule grules_last
     ==>
    [STOP]
enddefine;

;;; G_default---------------------------------------------------------

;;; utilities---------------------------------------------------------

;;; rules-------------------------------------------------------------

define :ruleset G_default;

rule grule_default_add
    [NOT MOTIVE == status active ==]
    [NOT MOTIVE motive [default] ==]
==>
    [MOTIVE motive [default] insistence 0.01 status sub]

rule grules_last
     ==>
    [STOP]
enddefine;

;;; G_condense_motives ------------------------------------------------

;;; utilities---------------------------------------------------------

;;; rules-------------------------------------------------------------

;;; If more than one motivator about *same* object in environment,
;;; then condense it into one motive.

define :ruleset G_condense_motives;

rule init_descriptor_table
    [NOT motive_table ==]
==>
    [LVARS Time]
    [POP11 sim_myself.trsim_clock -> Time; ]
    [motive_table time ?Time]
    [STOP]

rule condense_motives_add
    [motive_table == [?Name1 ??Mid1 ?Obj] == [?Name2 ??Mid2 ?Obj] ==]
        ;;; this will detect for two motives that concern
        ;;; the same object
    [WHERE (Name1 /= "condense") and (Name2 /= "condense") ]
    [NOT MOTIVE motive [condense [?Name1 ??Mid1 ?Obj] [?Name2 ??Mid2 ?Obj]] ==]
    [NOT noted same ?Name1 =]
    [NOT noted same = ?Name1]
    [NOT noted same ?Name2 =]
    [NOT noted same = ?Name2]
==>
    ;;; Create a condense motive
    [MOTIVE motive [condense [?Name1 ??Mid1 ?Obj] [?Name2 ??Mid2 ?Obj]]
        insistence 0.8 status sub]
    [noted same ?Name1 ?Name2]

rule clear_descriptor_table
    [motive_table == ?Time][ ->> Mtable ]
    [WHERE Time < sim_myself.trsim_clock ]
==>
;;;    [POP11 Mtable ==> ]
    [POP11 sim_myself.trsim_clock -> Time; ]
    [REPLACE 1 [motive_table time ?Time]]

rule create_descriptor_table
    [MOTIVE motive ?Descriptor == status ?Status ==]
    [WHERE Status /= "sub" ]
    [NOT noted motive ?Descriptor]
==>
    [PUSH ?Descriptor motive_table]
    [noted motive ?Descriptor]

rule condense_motives_last
==>
    [NOT noted motive ==]
    [NOT noted same ==]
    [STOP]

enddefine;

;;; trsim_filter_motives--------------------------------------------

;;; utilities-------------------------------------------------------

;;; rules-----------------------------------------------------------

define :ruleset trsim_filter_motives;

rule init_filter
    [NOT filter ==]
     ==>
    [filter 0.01]

rule init_count
    [NOT motive_count ==]
==>
    [motive_count 0]

rule filter_r1
    [MOTIVE == insistence ?Insist status sub ==]
        [ ->> Motive ]
    [filter ?Threshold]
    [WHERE Insist > Threshold ]
     ==>
    [MODIFY 1 status surfacing]
    [SAY [Surfacing -- ?Motive]]

rule filter_r2
    [MOTIVE == insistence ?Insist status suspended ==]
        [ ->> Motive ]
    [filter ?Threshold]
    [WHERE Insist <= Threshold ]
     ==>
    [MODIFY 1 status sub]
    [SAY [Diving -- ?Motive]]

rule filter_r3
    ;;; count number of surfaced motives
    ;;; Note: due to limit on cycles, 6 is maximum
    [MOTIVE motive ?Name == status ?Status ==]
    [WHERE Status /= "sub"]
    [NOT noted ?Name]
    [motive_count ?Count][->> Mcount]
==>
    [noted ?Name]
    [POP11 Count + 1 -> Count; ]
    [REPLACE ?Mcount [motive_count ?Count]]

rule filter_r4
    ;;; raise filter threshold if too many surfaced motives
    [motive_count ?Count]
    [WHERE Count > RL_max_motives ]
    [filter ?Threshold][ ->> Filter]
    [WHERE Threshold < 1 ]
==>
    ;;; RESOURCE LIMIT: Can only have 3 (RL_max_motives) management motives,
    ;;; e.g. 1 active, 2 suspended, or 2 active, 1 suspended
    ;;; If more than RL_max_motives then raise filter threshold
    [POP11 Threshold + 0.01 -> Threshold; ]
        ;;; bit like arousal level ?!
        ;;; If so, arousal isn't emotion, but is another mechanism
        ;;; for focusing attention.
    [REPLACE ?Filter [filter ?Threshold]]
    [SAY [RAISING filter threshold to ?Threshold]]

rule filter_r5
    ;;; lower filter threshold if not enough surfaced motives
    [motive_count ?Count]
    [WHERE Count < RL_max_motives ]
    [filter ?Threshold][ ->> Filter]
    [WHERE Threshold > 0 ]
==>
    [POP11 Threshold - 0.01 -> Threshold; ]
    [REPLACE ?Filter [filter ?Threshold]]
    [SAY [LOWERING filter threshold to ?Threshold]]

rule filter_rules_last
;;;    [motive_count ==][->> Mcount]
     ==>
;;;    [SAY ?Mcount]
    [NOT noted ==]
    [NOT motive_count ==]
    [STOP]
enddefine;

;;; trsim_get_active_motives ---------------------------------------

;;; NOTE: want to implement various motive management strategies
;;; as reasoning goals, with TRP-like plan-steps.

;;; utilities-------------------------------------------------------

;;; rules-----------------------------------------------------------

;;; Here, there will be some already active motives (which are
;;; currently being pursued), and perhaps a set of newly surfaced
;;; motives that need to be decided.

define :ruleset trsim_get_active_motives;

;;; Temporary kludge for testing

;;; Rules for immediately deciding newly surfaced motives

rule decide_surfaced_motives_r1
    [MOTIVE == status surfacing ==] [->> Motive]
==>
    [MODIFY ?Motive status suspended]

;;; from the set of motives of status E {suspended, active} need
;;; to pick out the active motive, which, currently, is the motive
;;; with the highest insistence

rule get_active_motive_r1
;;; if no active motive then choose a suspended motive to
;;; make active -- this may well be quickly displaced.
    [NOT MOTIVE == status active ==]
    [MOTIVE == status suspended ==][ ->> Motive]
==>
    [MODIFY ?Motive status active]
    [STOP]

rule get_active_motive_r2
;;; if no active motive or suspended motives,
;;; then make a subliminal motive active (raise
;;; its insistence too).
    [NOT MOTIVE == status active ==]
    [NOT MOTIVE == status suspended ==]
    [MOTIVE == insistence ?Insistence status sub ==][ ->> Motive ]
    [filter ?Threshold]
==>
    [LVARS newins]
    [POP11
        if Insistence <= Threshold then
            Threshold + 0.01 -> newins
        else
            Insistence -> newins
        endif;
    ]
    [MODIFY ?Motive insistence ?newins status active]
    [STOP]

rule get_active_motive_r3
    [MOTIVE motive ?Descriptor1 insistence ?Insist1 status suspended ==]
        [ ->> Motive1]
    [MOTIVE motive ?Descriptor2 insistence ?Insist2 status active ==]
        [ ->> Motive2]
    [NOT noted ?Descriptor1 with ?Descriptor2]
    [WHERE Insist1 > Insist2]
    ==>
    [MODIFY ?Motive1 status active]
    [MODIFY ?Motive2 status suspended]
    [noted ?Descriptor2 with ?Descriptor1]

rule get_active_motive_r4
    [MOTIVE motive ?Descriptor1 insistence ?Insist1 status suspended ==]
        [ ->> Motive1]
    [MOTIVE motive ?Descriptor2 insistence ?Insist2 status active ==]
        [ ->> Motive2]
    [NOT noted ?Descriptor1 with ?Descriptor2]
    [WHERE Insist1 < Insist2]
    ==>
    [noted ?Descriptor1 with ?Descriptor2]

;;; active motive should now be motive with highest insistence
;;; and there should only be *one* active motive

rule manage_motives_r1
    [MOTIVE motive [?Name ??Params] insistence ?Insist status active]
        [ ->> Motive ]
     ==>
    [DEL 1]
    [REPLACE 1 [MOTIVE motive [?Name ??Params]
                       insistence ?Insist
                       status active
                       plan [[decide] [get_plan]]
;;;                       plan [[get_plan]]
                       trp [] ]
    ]
    [SAY [Activated -- ?Motive]]

rule manage_motives_last
     ==>
    [NOT noted ==]
    [STOP]
enddefine;

;;; trsim_order_motives----------------------------------------------

;;; Here we take active and suspended motives and order them.
;;; Assumes cost measure already calculated by "decide" process.

;;; Types of cost are: easy, ok, hard, impossible

;;; impossible motives should have their insistence set to 0,
;;; i.e. be supressed.

;;; utilities---------------------------------------------------------

;;; rules-------------------------------------------------------------

define :ruleset trsim_order_motives;


rule order_motives_last
    ==>
    [STOP]

enddefine;

;;; trsim_runplan_rules---------------------------------------------

;;; utilities-------------------------------------------------------

;;; rules-----------------------------------------------------------

define :ruleset trsim_runplan_rules;

;;; For printing.
;;; will need to be changed if more than one active motive
rule set_active_motive_slot_r1
    [MOTIVE motive ?Descriptor == status active ==]
==>
    [POP11 Descriptor -> sim_myself.trsim_active_motive; ]

rule set_active_motive_slot_r2
    [NOT MOTIVE == status active ==]
==>
    [POP11 [] -> sim_myself.trsim_active_motive; ]

rule plan_done
    [MOTIVE == status active == plan [] ==]
        [ ->> Motive]
     ==>
    [SAY [Completed -- ?Motive]]
    [DEL 1]

rule trp_done
;;; [NOT MOTIVE == plan [] ==]
    [MOTIVE == status active == plan [?Goal ??Rest] trp [done ??This] ==]
    [WHERE Goal matches This ]
     ==>
    [MODIFY 1 plan [??Rest] trp []]

rule make_trp
    [MOTIVE == status active == plan [[??Goal] ==] trp [] ==]
     ==>
    [MODIFY 1 trp [??Goal]]

rule runTRP
    [MOTIVE == status active == plan [[?Goal ??Info] ==] trp [= ==] ==]
        [ ->> Motive ]
    [NOT noted ?Motive]
;
    [MODIFY 1 trp [?Goal ??Info]]
    [noted ?Motive]
    [POP11
        false -> prb_allrules;
        vars TR_program;
        TR_index --> [== ^Goal ?TR_program ==];
        prb_run(valof(TR_program), sim_myself.sim_data, false);
        true -> prb_allrules;
    ]

rule last_actor_rules
     ==>
    [NOT noted ==]
    [POP11 true -> prb_allrules; ]
    [STOP]
enddefine;

;;; SP_leave---------------------------------------------------

;;; Special plan step.
;;; Sets motive insistence to zero (therefore should never be
;;; active). Note: for generactivators that continually modify
;;; the insistence of their motivators, this could be a problem.
;;; But also has effect that the plan for this motivator never
;;; completes; therefore, motive will only be removed by the
;;; generactivator, and not by management.

;;; Currently used for dealing with frighteneds, where the frightened needs
;;; to be left in the room until it calms down.

;;; utilities--------------------------------------------------

;;; rules------------------------------------------------------

define :ruleset SP_leave;

rule SP_leave_r1
    [MOTIVE motive == status active == trp [leave] ==]
        ==>
    [MODIFY 1 insistence 0 status sub trp [stop]]

rule SP_leave_last
        ==>
    [STOP]

enddefine;

;;; SP_get_plan ----------------------------------------------

;;; Gets plan for a motivator

;;; utilities ------------------------------------------------

;;; rules ----------------------------------------------------

define :ruleset SP_get_plan;

rule SP_get_plan_low_charge
    [MOTIVE motive [low_charge ??Params] == status active == trp [get_plan] ==]
        ==>
    [MODIFY 1 plan [[charge_object ??Params 200]] trp [] ]
    [STOP]

rule SP_get_plan_bug_dead
    [MOTIVE motive [bug_dead ??Params] == status active == trp [get_plan] ==]
        ==>
    [MODIFY 1 plan [[dismiss_object ??Params]] trp [] ]
    [STOP]

rule SP_get_plan_is_frightened
    [MOTIVE motive [is_frightened ??Params] == status active == trp [get_plan] ==]
        ==>
    ;;; need something more intelligent here as
    ;;; depending on the status of the world, will
    ;;; need to construct different plans. E.g.,
    ;;; a bug may become a frightened when it is already
    ;;; in a shut room -- this means that the actor
    ;;; will continually try to grab it without success
    ;;; if it follows the plan below.

    ;;; Also, need a superior plan representation and
    ;;; mechanism. Note that the plan below may be
    ;;; interrupted at anytime. Therefore, it is
    ;;; possible that the actor won't be holding the
    ;;; bug when it wishes to enter the room, or the
    ;;; door has been closed by a seperate motive but
    ;;; the plan below assumes that it is open ... etc.

    ;;; Also agent can shut frightened bug in with
    ;;; thugs -- disaster!

    [MODIFY 1 plan
        [
            [make_wall %front(ap_room_loc)% %back(ap_room_loc) +
                ap_offset% 0 north]
            [make_wall %front(ap_room_loc) + ap_offset%
                %back(ap_room_loc)% 90 east]
            [make_wall %front(ap_room_loc)% %back(ap_room_loc) - ap_offset%
                0 south]
            [make_wall %front(ap_room_loc) - ap_offset - 4%
                %back(ap_room_loc)% 65 west]
            [grab_object ??Params]
            [amble %front(ap_room_loc) - 20% %back(ap_room_loc)%]
            [drop ??Params]
            [open_door]
            [grab_object ??Params]
            [amble %destpair(ap_room_loc)%]
            [drop ??Params]
            [shut_door]
            [leave] ;;; special term: motive never
            ;;; completes; therefore, is only
            ;;; removed by generactivator process

            ;;; need to let out frightened babies when they
            ;;; have calmed down
        ]
        trp []
    ]
    [STOP]

rule SP_get_plan_visit
    [MOTIVE motive [visit ??Params] == status active == trp [get_plan] ==]
    ==>
    [MODIFY 1 plan [[drop] [goto_object ??Params]] trp [] ]
    [STOP]

rule SP_get_plan_near_ditch
    [MOTIVE motive [near_ditch ??Params] == status active == trp [get_plan] ==]
    ==>
    [MODIFY 1 plan [[put_safe ??Params]] trp [] ]
    [STOP]

rule SP_get_plan_no_maginot
    [MOTIVE motive [no_maginot ??Params] == status active == trp [get_plan] ==]
    ==>
    ;;; will need to employ some reasoning in order to
    ;;; swap between maginot line and room, i.e. need to
    ;;; explicitly destory a structure
    [MODIFY 1 plan [[make_wall 15 60 0 first]
            [make_wall 40 60 0 second]
            [make_wall 65 60 0 third]
            [make_wall 90 60 0 fourth]]
            trp []
    ]
    [STOP]

rule SP_get_plan_below_maginot
    [MOTIVE motive [below_maginot ??Params] == status active == trp [get_plan] ==]
    ==>
    [MODIFY 1 plan [[take_object ??Params exit1]] trp [] ]
    [STOP]

rule SP_get_plan_default
    [MOTIVE motive [default ??Params] == status active == trp [get_plan] ==]
    ==>
    [MODIFY 1 plan [[drop] [search]] trp [] ]
    [STOP]

rule SP_condense_motives
    [MOTIVE motive [condense ??Params] == status active == trp [get_plan] ==]
==>
    [MODIFY 1 plan [[condense ??Params]] trp [] ]
    [STOP]

rule SP_get_plan_last
    ==>
    [SAY 'Unknown motive descriptor type']
    [POP11 interrupt(); ]
    [STOP]

enddefine;

;;; SP_condense_motives ----------------------------------------------

/*
    Types of motive are:
    low_charge
    below_maginot
    bug_dead
    is_frightened
    visit
    near_ditch
    no_maginot
    default

(list may not be complete)

Do not destroy single motives, but sub them and
set insistence to 0.

Ensure that they are removed when compound motive
succeeds.

Possible combinations are:

low_charge + near_ditch = low_charge with higher insistence

*/

;;; utilities ------------------------------------------------

;;; rules ----------------------------------------------------

define :ruleset SP_condense_motives;

rule SP_condense_motives_r1
    [MOTIVE motive == status active ==
        trp [condense [?Name1 ??Params1] [?Name2 ??Params2]] ==]
    [->> Motive]
    [WHERE (Name1 = "low_charge" and Name2 = "near_ditch")
        or (Name1 = "near_ditch" and Name2 = "low_charge") ]
==>
    [SAY ?Motive]
    ;;; Destroy condense motive
    ;;; sub near_ditch motive, and ins = 0
    ;;; increase urgency of low_charge (insistence will be
    ;;; continusouly modified).


rule SP_condense_motives_last
    ==>
    [STOP]

enddefine;

;;; SP_decide-------------------------------------------------

;;; utilities-------------------------------------------------

;;; rules-----------------------------------------------------

;;; Temporary kludge to avoid considerations of
;;; (a) reward for successful achievement of motive
;;; (b) cost/benefit analyses
;;; (c) consideration of time
;;; (d) opportunities and risks
;;; (e) urgency and importance etc.

;;; Instead will determine a priority, based on
;;; insistence level and ease of achieving motive.
;;; ALSO, if motive is definitely unachievable, then
;;; REJECT it, i.e. lower its insistence level to 0
;;; (This will hopefully produce perturbant states).

/*
    Types of motive and (ad hoc) cost of achieving motive

    G_low_charge
        Low if knows where gas1 and bug is.
        Ok if knows where bug or gas1 is.
        High if knows nothing.
    G_bug_dead
        Low if knows where exit1 and bug is.
        Ok if knows where exit1 or bug is.
        High if knows nothing. (Unlikely).
    G_is_frightened
        Low if knows where 4 bars are
                        where bug is
        Ok if knows where bug is
        High if knows nothing.
    G_patrol_ditches
        Low if near ditch
        High otherwise
    G_near_ditch
        Low if knows where bug is
        Mid otherwise
    G_no_maginot
        Low if knows where 4 bars are
        Ok if knows 2 or more bars
        High otherwise
    G_below_maginot
        Low if knows where bug is
        High otherwise
    G_default
        High
    G_condense_motives
        Low
*/

define :ruleset SP_decide;

rule decide_low_charge_low
    [MOTIVE motive [low_charge ?Obj] == status ?Status == trp [decide] == cost ==]
    [WHERE Status = "suspended" or Status = "active" ]
    [belief == name ?Obj ==]
    [belief == name gas1 ==]
==>
    [MODIFY 1 trp [done decide] cost low]

rule decide_low_charge_ok
    [MOTIVE motive [low_charge ?Obj] == status ?Status == trp [decide] == cost ==]
    [WHERE Status = "suspended" or Status = "active" ]
    [OR [belief == name ?Obj ==] [belief == name gas1 ==]]
==>
    [MODIFY 1 trp [done decide] cost ok]

rule decide_low_charge_high
    [MOTIVE motive [low_charge ?Obj] == status ?Status == trp [decide] == cost ==]
    [WHERE Status = "suspended" or Status = "active" ]
==>
    [MODIFY 1 trp [done decide] cost high]

rule init_cost
    [MOTIVE motive ??Bit1 status ?Status ??Bit2 trp [decide] ??Bit3][->> Motive]
    [WHERE Status = "suspended" or Status = "active" ]
    [WHERE not(Motive matches [== cost = ==]) ]
==>
    [REPLACE ?Motive [MOTIVE motive ??Bit1 status ?Status ??Bit2
                    trp [stop] ??Bit3 cost undef]]

rule catch_rest
    [MOTIVE motive == status ?Status == trp [decide] == cost undef ==]
    [WHERE Status = "suspended" or Status = "active" ]
==>
    [MODIFY 1 trp [done decide]]

rule SP_decide_last
==>
    [STOP]

enddefine;

;;; TRP_goto--------------------------------------------------

;;; utilities-------------------------------------------------

define procedure close_enough(x1, y1, x2, y2) -> bool;
    ;;; changed to use four coordinates
lvars loc1, loc2, bool;
lconstant proximity = 0.25; ;;; for bar width = 0

    abs(x1 - x2) < proximity and
            abs(y1 - y2) < proximity -> bool
enddefine;

/*
    close_enough([10 10], [20 20]) =>
    close_enough([10 10], [10.4 10.9]) =>

    arcsin(10 / -20) =>
*/

;;; use sim_angle_xy(x, y) -> degrees  ;
;;; from lib sim_geom. Uses arctan2

define procedure get_heading_to_point(fx, fy, tx, ty) -> heading;
    sim_heading_from(fx, fy, tx, ty) -> heading;
enddefine;

/*
    get_heading_to_point(conspair(80,80), conspair(80,90)) =>
    get_heading_to_point(conspair(80,80), conspair(90,90)) =>
    get_heading_to_point(conspair(80,80), conspair(90,80)) =>
    get_heading_to_point(conspair(80,80), conspair(90,70)) =>
    get_heading_to_point(conspair(80,80), conspair(70,80)) =>
    get_heading_to_point(conspair(80,80), conspair(70,70)) =>
    get_heading_to_point(conspair(80,80), conspair(80,70)) =>
    get_heading_to_point(conspair(80,80), conspair(70,90)) =>
*/

define procedure course(fx, fy, tx, ty) -> heading;
    ;;; this really duplicates get_heading_to_point
    sim_heading_from(fx, fy, tx, ty) -> heading;
enddefine;

;;; rules-----------------------------------------------------------
;;; note: need to turn prb_allrules off when running a TR program

define :ruleset TRP_goto;

rule TR_goto_r1
    [MOTIVE motive == status active == trp [goto ?X ?Y] ==]
    [WHERE
        lvars (xself, yself) = destpair(sim_myself.trsim_location);
        ;;; should probably use a tolerance
        (xself = X) and (yself = Y)]
     ==>
    ;;; trp achieved
    [MODIFY 1 trp [done goto ?X ?Y]]
    [STOP]

rule TR_goto_r2
    [MOTIVE motive == status active == trp [goto ?X ?Y] ==]
    [WHERE
        heading_close_enough(
        course(destpair(sim_myself.trsim_location), X, Y),
        sim_myself.trsim_heading)
    ]
     ==>
    [LVARS
        [ssp =
            min
            (sim_distance_from(destpair(sim_myself.trsim_location), X,Y),
            sim_myself.trsim_max_speed) ]
    ]
    [closure setspeed ?ssp]
    [closure move]
    [MODIFY 1 trp [stop]]
    [STOP]

rule TR_goto_r3
    [MOTIVE motive == status active == trp [goto ?X ?Y] ==]
     ==>
    [LVARS ch_heading ssp]
    [POP11
        lvars (xself, yself) = destpair(sim_myself.trsim_location);
        course(xself, yself, X, Y) -> ch_heading;
        sim_distance_from(xself, yself, X, Y) -> ssp;
        min(ssp, sim_myself.trsim_max_speed) -> ssp
    ]
    [closure rotate ?ch_heading]
    [closure setspeed ?ssp]
    [closure move]
    [MODIFY 1 trp [stop]]
    [STOP]

rule TR_goto_last
    [MOTIVE motive == status active == trp [stop] ==]
     ==>
    [STOP]
enddefine;

;;; TRP_amble-------------------------------------------------

;;; utilities-------------------------------------------------------

define procedure return_new_point(xloc, yloc, dir, extension) -> (newx, newy);

        xloc + extension * cos(dir) -> newx;
        yloc + extension * sin(dir) -> newy;

enddefine;

/*
    return_new_point([0 0], 200, 10) =>
    return_new_point([50 50], 350, 18.5) =>
    return_new_point([0 0], 250, 10) =>
    return_new_point([0 0], 30, 10) =>
*/

define procedure clear_path(x1, y1, x2, y2, apolygon) -> bool;
lvars loc1, loc2, apolygon, bool, hitpoint;

    collision_detect(x1, y1, x2, y2, apolygon) -> bool -> hitpoint;
    if sim_distance_from(x1, y1, explode(hitpoint)) <= 0.0001 then
        false -> bool
    endif;
    not(bool) -> bool;

enddefine;

define :method is_bar_in_way( a:trsim_actor_agent, sense_list, xloc, yloc )
        -> inway;
    ;;; returns false if no bar in the way
lvars a, xloc, yloc, item, x, y, bool, hitpoint, newdist, inway = false;
vars the_poly;

    destpair(a.trsim_location) -> (x,y);
    for item in sense_list do
        item --> [== poly_space ?the_poly ==];
        ;;; something wrong here. This test does not depend on the loop
        if x = xloc and y = yloc then
            false -> bool
        else
            collision_detect(x, y, xloc, yloc, the_poly) -> bool -> hitpoint;
            if sim_distance_from(x, y, explode(hitpoint) ) <= 0.0001 then
                false -> bool
            endif;
        endif;
        if bool then true -> inway; quitloop; endif;
    endfor;

enddefine;

define :method return_clear_headings( a:trsim_actor_agent, xloc, yloc, dir )
        -> ch_list;
    ;;; two coords rather than loc
lvars a, ch_list, dir, inway, item, xnew, ynew,
        (x, y) = destpair(a.trsim_location),
        shift = dir * 10,
        odist = sim_distance_from(x, y, xloc, yloc),
        heading = get_heading_to_point(x, y, xloc, yloc),
        angular_distance = 0,
        sense_list = [% for item in (a.sim_data)("new_sense_data") do
                            if (item matches [== type bar status alive
                                    == held false ==])
;;;                                    or (item matches [== type wall ==])
                                    then
                                item
                            endif;
                        endfor; %]; ;;; construct list of sensed obstacles

    rotate_angle(heading, -shift) -> heading;
    [%
        repeat (180 div abs(shift)) times
            rotate_angle(heading, shift) -> heading;
            angular_distance + 1 -> angular_distance;
            return_new_point(x, y, heading, odist) -> (xnew, ynew);
            is_bar_in_way(a, sense_list, xnew, ynew) -> inway;
            unless inway then
                ;;; nothing inway
                conspair(heading, angular_distance);
                ;;; goto cheat;
                quitloop();
            endunless;
        endrepeat;
        cheat:
    %] -> ch_list;

enddefine;

define :method new_point( a:trsim_actor_agent, xloc, yloc, way ) -> (newx, newy);
;;; Slight cheat, as this computation should be done declaratively
;;; Need to find the nearest heading to current heading that
;;; is clear; failing that, give error as agent is boxed in.
lvars a, xloc, yloc, newx, newy, clear_headings_list, item, way, dir
        min_angle = 720, best_heading = false,
        (x,y) = destpair(a.trsim_location),
        odist = sim_distance_from(x, y, xloc, yloc);

    if way = "left" then 1 -> dir else -1 -> dir endif;
    return_clear_headings(a, xloc, yloc, dir) -> clear_headings_list;
    if null(clear_headings_list) then
        false ->> newx -> newy;
    else
        for item in clear_headings_list do
            if back(item) < min_angle then
                back(item) -> min_angle;
                front(item) -> best_heading;
            endif;
        endfor;
        unless best_heading then
            [ERROR: No best heading] =>
            interrupt();
        endunless;
        return_new_point(x, y, best_heading, odist) -> (newx, newy);
    endif;

enddefine;

;;; rules-----------------------------------------------------------
;;; includes constraints on recursive calls

/*
define :rule TR_recursion_limit in TRP_amble
    ;;; This recursion check won't work if there is noise
    ;;; in choice of movement direction.
    [trp == [amble ?X ?Y] == [amble ?X ?Y] ==]
    ;
    [VARS newx newy]
    [POP11
        random(world_size - 10.0) + 5 -> newx;
        random(world_size - 10.0) + 5 -> newy;
    ]
    [SAY 'Amble recursion limit exceeded']
    [DEL 1]
    [trp [goto ?newx ?newy]]
    [POP11
        prb_run(TRP_goto, sim_myself.sim_data, false);
    ]
    [STOP]
enddefine;
*/

define :ruleset TRP_amble;

rule TR_amble_r1
    [MOTIVE motive == status active == trp [amble ?X ?Y] ==]
    [WHERE
        lvars (xself, yself) = destpair(sim_myself.trsim_location);
        (xself = X) and (yself = Y) ]
        ==>
    ;;; trp achieved
    [MODIFY 1 trp [done amble ?X ?Y]]
    [STOP]

rule TR_amble_r2
    [MOTIVE motive == status active == trp [amble ?X ?Y] ==]
    [NOT_EXISTS
        [new_sense_data == type bar status alive == held false ==
            poly_space ?Polygon ==]
        [WHERE
            not(clear_path(
            destpair(sim_myself.trsim_location), X, Y, Polygon))]
    ]
        ==>
    ;;; call to TRP_goto
    [MODIFY 1 trp [goto ?X ?Y]]
    [POP11
        prb_run(TRP_goto, sim_myself.sim_data, false);
    ]
    [STOP]

rule TR_amble_r3
    [MOTIVE motive ??First status active ??Mid trp [amble ?X ?Y] ??End]
    [ ->> Motive ]
    [strategy ?Way]
    [LVARS newx newy]
    [POP11
        new_point(sim_myself, X, Y, Way) -> (newx, newy);
        if newx = false then
            ;;; blocked in
            sim_delete_data([strategy =], sim_myself.sim_data);
            if Way == "right" then
                sim_add_data([strategy left], sim_myself.sim_data);
                "left" -> Way;
            else
                sim_add_data([strategy right], sim_myself.sim_data);
                "right" -> Way;
            endif;
            new_point(sim_myself, X, Y, Way) -> (newx, newy);
            if newx = false then
                ;;; Could do a number of things:
                ;;;     1. reduce length of clear path required (dist)
                ;;;     2. new insistent plan to get out of trap,
                ;;;         with a special plan,
                ;;;     3. random amble,
                [ERROR agent boxed in] =>
                sim_myself.trsim_location =>
                interrupt();
            endif;
        endif;
    ]
    [WHERE (newx /= X) or (newy /= Y) ]
        ==>
    [REPLACE ?Motive [MOTIVE motive ??First status active ??Mid
        trp [amble ?newx ?newy] ??End]]
    [POP11
        prb_run(TRP_amble, sim_myself.sim_data, false);
    ]
    [STOP]

rule TR_amble_r3_5
    [MOTIVE motive == status active == trp [amble ?X ?Y] ==]
    [NOT strategy =]
        ==>
    [strategy left]
    [MODIFY 1 trp [stop]]
    [STOP]

rule TR_amble_r4
    [MOTIVE motive == status active == trp [amble ?X ?Y] ==]
        ==>
    [LVARS newx newy]
    [POP11
        random(world_size - 10.0) + 5 -> newx;
        random(world_size - 10.0) + 5 -> newy;
    ]
    [SAY 'Recursion trapped']
    [MODIFY 1 trp [goto ?newx ?newy]]
    [POP11
        prb_run(TRP_goto, sim_myself.sim_data, false);
    ]
    [STOP]

rule TR_amble_last
    [MOTIVE motive == status active == trp [stop] ==]
        ==>
    [STOP]
enddefine;

;;; TRP_goto_object -----------------------------------------------

;;; utilities ---------------------------------------------------------

;;; rules -------------------------------------------------------------

define :ruleset TRP_goto_object;

rule TR_goto_object_r1
    [MOTIVE motive == status active == trp [goto_object ?Obj] ==]
    [new_sense_data == name ?Obj == x ?X y ?Y ==]
    [WHERE
        close_enough(destpair(sim_myself.trsim_location), X, Y) ]
     ==>
    [MODIFY 1 trp [done goto_object ?Obj]]
    [STOP]

rule TR_goto_object_r2
    [MOTIVE motive == status active == trp [goto_object ?Obj] ==]
    [new_sense_data == name ?Obj == x ?X y ?Y ==]
     ==>
    ;;; amble to object's midpoint
    [LVARS dist hding newcds]
    [POP11
        lvars (xself, yself) = destpair(sim_myself.trsim_location);
        sim_distance_from(xself, yself, X,Y) -> dist;
        dist / bounce_ratio -> hding;
        dist - hding -> dist;
        get_heading_to_point(xself, yself, X,Y) -> hding;
        return_new_point(xself, yself, hding, dist) -> (X,Y);
    ]
    [MODIFY 1 trp [amble ?X ?Y]]
    [POP11
        prb_run(TRP_amble, sim_myself.sim_data, false);
    ]
    [STOP]

rule TR_goto_object_r3
    [MOTIVE motive == status active == trp [goto_object ?Obj] ==][ ->> M_s]
    [belief == name ?Obj == x ?X y ?Y ==]
     ==>
    ;;; amble to last time agent saw object's midpoint
    [POP11
        lvars
            dist, hding,
            (xself, yself) = destpair(sim_myself.trsim_location);

        sim_distance_from(xself, yself, X,Y) -> dist;

        dist / bounce_ratio -> hding;
        dist - hding -> dist;
        get_heading_to_point(xself, yself, X,Y) -> hding;
        return_new_point(xself, yself, hding, dist) -> (X, Y);
    ]
    [MODIFY ?M_s trp [amble ?X ?Y]]
    [POP11
        prb_run(TRP_amble, sim_myself.sim_data, false);
    ]
    [STOP]

rule TR_goto_object_r4
    [MOTIVE motive == status active == trp [goto_object ?Obj] ==]
     ==>
    ;;; search for object, as agent has no information about it
    [MODIFY 1 trp [search]]
    [POP11
        prb_run(TRP_search, sim_myself.sim_data, false);
    ]
    [STOP]

rule TR_goto_object_last
    [MOTIVE motive == status active == trp [stop] ==]
     ==>
    [STOP]
enddefine;

;;; TRP_grab_object -----------------------------------------------

;;; utilities ---------------------------------------------------------

;;; rules -------------------------------------------------------------

define :ruleset TRP_grab_object;

rule TR_grab_object_r1
    [MOTIVE motive == status active == trp [grab_object ?Obj] ==]
    [held ?Obj]
     ==>
    ;;; trp achieved
    [MODIFY 1 trp [done grab_object ?Obj]]
    [STOP]

rule TR_grab_object_r2
    [MOTIVE motive == status active == trp [grab_object ?Obj] ==]
    [new_sense_data == name ?Obj == x ?X y ?Y ==]
    [WHERE
        close_enough(destpair(sim_myself.trsim_location), X, Y) ]
    [NOT held ==]
     ==>
    ;;; grab the object
    [closure grab_object ?Obj]
    [MODIFY 1 trp [stop]]
    [STOP]

rule TR_grab_object_r3
    [MOTIVE motive == status active == trp [grab_object ?Obj] ==]
    [NOT held ==]
     ==>
    [MODIFY 1 trp [goto_object ?Obj]]
    [POP11
        prb_run(TRP_goto_object, sim_myself.sim_data, false);
    ]
    [STOP]

rule TR_grab_object_r4
    [MOTIVE motive == status active == trp [grab_object ?Obj] ==]
    [held ?Obj2]
     ==>
    [MODIFY 1 trp [drop ?Obj2]]
    [POP11
        prb_run(TRP_drop, sim_myself.sim_data, false);
    ]
    [STOP]
enddefine;

;;; TRP_take_object -----------------------------------------------

;;; utilities ---------------------------------------------------------

;;; rules -------------------------------------------------------------

define :ruleset TRP_take_object;

rule TR_take_object_r1
    [MOTIVE motive == status active == trp [take_object ?Obj ?Obj2] ==]
    [new_sense_data == name ?Obj == x ?X y ?Y ==]
    [new_sense_data == name ?Obj2 == x ?XX y ?YY ==]
    [WHERE close_enough(X, Y, XX, YY) ]
    [held ?Obj]
     ==>
    [MODIFY 1 trp [done take_object ?Obj ?Obj2]]
    [STOP]

rule TR_take_object_r2
    [MOTIVE motive == status active == trp [take_object ?Obj ?Obj2] ==]
    [held ?Obj]
     ==>
    [MODIFY 1 trp [goto_object ?Obj2]]
    [POP11
        prb_run(TRP_goto_object, sim_myself.sim_data, false);
    ]
    [STOP]

rule TR_take_object_r3
    [MOTIVE motive == status active == trp [take_object ?Obj ?Obj2] ==]
     ==>
    [MODIFY 1 trp [grab_object ?Obj]]
    [POP11
        prb_run(TRP_grab_object, sim_myself.sim_data, false);
    ]
    [STOP]

rule TR_take_object_last
    [MOTIVE motive == status active == trp [stop] ==]
     ==>
    [STOP]
enddefine;

;;; TRP_drop -----------------------------------------------

;;; utilities ----------------------------------------------------

define :method inside_poly( polygon:poly, x, y ) -> bool;
lvars polygon, x, y, bool;
lvars x1, y1, x2, y2, x3, y3, x4, y4, hit, loc;
lconstant ll = world_size;
    ;;; x,y are coords. Check if they lie within poly.
    false -> bool;
    x + ll -> x1; y -> y1;
    x -> x2; y - ll -> y2;
    x - ll -> x3; y -> y3;
    x -> x4; y + ll -> y4;
    collision_detect(x, y, x1, y1, polygon) -> hit -> loc;
    if hit then
        collision_detect(x, y, x2, y2, polygon) -> hit -> loc;
        if hit then
            collision_detect(x, y, x3, y3, polygon) -> hit -> loc;
            if hit then
                collision_detect(x, y, x1, y1, polygon) -> hit -> loc;
                if hit then true -> bool; endif;
            endif;
        endif;
    endif;
enddefine;

/*
    ditch1.bar_current_poly =>
    bug1 =>
    vars x, y;
    destpair(bug1.trsim_location) -> (x,y); x => y =>
    conspair(4.9,30) -> bug1.trsim_location;
    draw_agent(bug1);
    inside_poly(ditch1.bar_current_poly, front(bug1.trsim_location),
        back(bug1.trsim_location)) =>
*/

;;; rules --------------------------------------------------------

define :ruleset TRP_drop;

rule TR_drop_r1
    [MOTIVE motive == status active == trp [drop ?Obj] ==]
    [NOT held ?Obj]
     ==>
    [MODIFY 1 trp [done drop ?Obj]]
    [STOP]

rule TR_drop_r2
    [MOTIVE motive == status active == trp [drop] ==]
    [NOT held ==]
     ==>
    [MODIFY 1 trp [done drop]]
    [STOP]

rule TR_drop_r3
    [MOTIVE motive == status active == trp [drop ==] ==]
    [held ?Obj]
    [NOT_EXISTS
        [belief == name ?Obj == type bug == x ?X y ?Y ==]
        [new_sense_data == type ditch == x ?X1 y ?Y1 == poly_space ?Polygon ==]
        [WHERE
            inside_poly(Polygon, X, Y)
                or is_near_ditch(Polygon, X, Y, X1, Y1, near_ditch_proximity) ]
    ]   ;;; this prevents agent dropping a bug when it is over a ditch
     ==>
    [closure drop ?Obj]
    [MODIFY 1 trp [stop]]
    [STOP]

rule TR_drop_r5
    ;;; agent must be over a ditch and wants to drop a bug
    [MOTIVE motive == status active == trp [drop ==] ==]
    ==>
    [LVARS loc]
    [POP11 world_size div 2 -> loc; ]
    [MODIFY 1 trp [amble ?loc ?loc]]
    [POP11
        prb_run(TRP_amble, sim_myself.sim_data, false);
    ]
    [STOP]

rule TR_drop_last
    [MOTIVE motive == status active == trp [stop] ==]
     ==>
    [STOP]
enddefine;

;;; TRP_search -----------------------------------------------

;;; utilities ------------------------------------------------------

;;; rules ----------------------------------------------------------

define :ruleset TRP_search;

rule TR_search_r1
    [MOTIVE motive == status active == trp [search] ==]
    [search_state]
     ==>
    ;;; search of room done
    [MODIFY 1 trp [done search]]
    [NOT search_state ==]
    [STOP]

rule TR_search_r2
    [MOTIVE motive == status active == trp [search] ==]
    [search_state [?X ?Y] ==]
    [WHERE
        lvars (x,y) = destpair(sim_myself.trsim_location);
        ;;; should probably use a tolerance ? (A.S.)
        (x = X) and (y = Y) ]
     ==>
    [POP search_state]
    [MODIFY 1 trp [stop]]
    [STOP]

rule TR_search_r3
    [MOTIVE motive == status active == trp [search] ==]
    [search_state [?X ?Y] ==]
     ==>
    [MODIFY 1 trp [amble ?X ?Y]]
    [POP11
        prb_run(TRP_amble, sim_myself.sim_data, false);
    ]
    [STOP]

rule TR_search_r4
    [MOTIVE motive == status active == trp [search] ==]
     ==>
    [LVARS search_list]
    [POP11
        lvars r1, r2, w_5;
        random(world_size - 20) + 10 -> r1;
        random(world_size - 20) + 10 -> r2;
        world_size - 5 -> w_5;
        [ [5 ^w_5] [^w_5 ^w_5] [^w_5 5] [5 5] [5 ^w_5] [^r1 ^r2] ]
            -> search_list;
        shuffle(search_list) -> search_list;
    ]
    [NOT search_state ==]
    [search_state ??search_list ]
    [MODIFY 1 trp [stop]]
    [STOP]

rule TR_search_last
    [MOTIVE motive == status active == trp [stop] ==]
     ==>
    [STOP]
enddefine;

;;; TRP_place_bar ------------------------------------------------------

;;; rules --------------------------------------------------------------

define :ruleset TRP_place_bar;

rule TR_place_bar_r1
    ;;; bar is in correct place, has correct heading and
    ;;; is not held.
    [MOTIVE motive == status active == trp [place_bar ?Obj ?X ?Y ?Heading] ==]
    [new_sense_data == name ?Obj type bar status alive == x ?XX y ?YY
        == held false == heading ?Hheading ==]
    [WHERE close_enough(X, Y, XX, YY) and
        is_right_orientation(Heading, Hheading) ]
     ==>
    [MODIFY 1 trp [done place_bar ?Obj ?X ?Y ?Heading]]
    [STOP]

rule TR_place_bar_r2
    ;;; bar is in correct place, has correct heading, and is held
    [MOTIVE motive == status active == trp [place_bar ?Obj ?X ?Y ?Heading] ==]
    [new_sense_data == name ?Obj type bar status alive == x ?XX y ?YY
        == held true == heading ?Hheading ==]
    [WHERE close_enough(X, Y, XX, YY) and
        is_right_orientation(Heading, Hheading) ]
    [held ?Obj]
     ==>
    [MODIFY 1 trp [drop ?Obj]]
    [POP11
        prb_run(TRP_drop, sim_myself.sim_data, false);
    ]
    [STOP]

rule TR_place_bar_r3
    ;;; bar is in correct place and is held
    [MOTIVE motive == status active == trp [place_bar ?Obj ?X ?Y ?Heading] ==]
    [new_sense_data == name ?Obj type bar status alive == x ?XX y ?YY
        == held true == Heading ?Hheading ==]
    [WHERE close_enough(X, Y, XX, YY) ]
    [held ?Obj]
     ==>
    [LVARS [amnt = difference_in_heading(Hheading, Heading) ]]
    [closure rotate_bar ?Obj ?amnt]
    [MODIFY 1 trp [stop]]
    [STOP]

rule TR_place_bar_r4
    ;;; bar has correct heading and is held
    [MOTIVE motive == status active == trp [place_bar ?Obj ?X ?Y ?Heading] ==]
    [new_sense_data == name ?Obj type bar status alive == x ?XX y ?YY
        == held true == heading ?Hheading ==]
    [WHERE is_right_orientation(Heading, Hheading) ]
    [held ?Obj]
     ==>
    [MODIFY 1 trp [amble ?X ?Y]]
    [POP11
        prb_run(TRP_amble, sim_myself.sim_data, false);
    ]
    [STOP]

rule TR_place_bar_r5
    ;;; bar is held
    [MOTIVE motive == status active == trp [place_bar ?Obj ?X ?Y ?Heading] ==]
    [new_sense_data == name ?Obj type bar status alive ==
        held true == heading ?Hheading ==]
    [held ?Obj]
     ==>
    [LVARS amnt]
    [POP11 difference_in_heading(Hheading, Heading) -> amnt; ]
    [closure rotate_bar ?Obj ?amnt]
    [MODIFY 1 trp [amble ?X ?Y]]
    [POP11
        prb_run(TRP_amble, sim_myself.sim_data, false);
    ]
    [STOP]

;;; if last rule fires, then preconditions for this tr program
;;; have not been met and it fails.

rule TR_place_bar_last
    [MOTIVE motive == status active == trp [stop] ==]
     ==>
    [STOP]
enddefine;

;;; TRP_grab_wall -----------------------------------------------

;;; utilities ---------------------------------------------------------

;;; rules -------------------------------------------------------------

define :ruleset TRP_grab_wall;

rule TR_grab_wall_r1
    ;;; a bar object is held and it isn't part of a room already
    [MOTIVE motive == status active == trp [grab_wall] ==]
    [held ?Obj]
    [NOT belief name = wall is ?Obj ==]
    [belief == name ?Obj type bar == held true ==]
     ==>
    [MODIFY 1 trp [done grab_wall]]
    [NOT candidate_wall is ==]
    [STOP]

rule TR_grab_wall_r2
    ;;; an object is held that is part of a room already
    [MOTIVE motive == status active == trp [grab_wall] ==]
    [held ?Obj]
    [belief name = wall is ?Obj ==]
     ==>
    [MODIFY 1 trp [drop ?Obj]]
    [POP11
        prb_run(TRP_drop, sim_myself.sim_data, false);
    ]
    [STOP]

rule TR_grab_wall_r3
    ;;; an object is held that is not a bar
    [MOTIVE motive == status active == trp [grab_wall] ==]
    [held ?Obj]
    [NOT belief == name ?Obj type bar ==]
     ==>
    [MODIFY 1 trp [drop ?Obj]]
    [POP11
        prb_run(TRP_drop, sim_myself.sim_data, false);
    ]
    [STOP]

rule TR_grab_wall_r4
    ;;; candidate wall chosen
    [MOTIVE motive == status active == trp [grab_wall] ==]
    [belief name candidate_wall is ?Obj]
    [NOT belief name = wall is ?Obj ==]
     ==>
    [MODIFY 1 trp [grab_object ?Obj]]
    [POP11
        prb_run(TRP_grab_object, sim_myself.sim_data, false);
    ]
    [STOP]

rule TR_grab_wall_r5
    [MOTIVE motive == status active == trp [grab_wall] ==]
    [belief name candidate_walls are ??Objs]
    [WHERE not(null(Objs)) ]
     ==>
    [NOT belief name candidate_walls are ==]
    [LVARS thewall]
    [POP11
        lvars item, thedist,
            mindist = world_size * 2,
            (xself, yself) = destpair(sim_myself.trsim_location);
        for item in Objs do
            sim_distance_from(xself, yself,
                destpair(trsim_location(valof(item)))) -> thedist;
            if thedist < mindist then
                thedist -> mindist;
                item -> thewall;
            endif;
        endfor;
    ]
    [NOT belief name candidate_wall is ==]
    [belief name candidate_wall is ?thewall]
    [MODIFY 1 trp [stop]]
    [STOP]

rule TR_grab_wall_r6
    [MOTIVE motive == status active == trp [grab_wall] ==]
    [NOT belief name candidate_walls are ==]
     ==>
    [belief name candidate_walls are]
    [MODIFY 1 trp [stop]]
    [STOP]

rule TR_grab_wall_r7
    ;;; agent can sense (or believes) a bar that is not part of a room
    [MOTIVE motive == status active == trp [grab_wall] ==]
    [OR [new_sense_data == name ?Obj type bar status alive == id 'bar'
            held false ==]
        [belief == name ?Obj type bar status alive == id 'bar'
            held false ==] ]
    [NOT belief name = wall is ?Obj ==]
    [belief name candidate_walls are ??Rest] [ ->> Candidates]
     ==>
    [REPLACE ?Candidates [belief name candidate_walls are ??Rest ?Obj]]
    [MODIFY 1 trp [stop]]
    [STOP]

rule TR_grab_wall_r8
    [MOTIVE motive == status active == trp [grab_wall] ==]
     ==>
    [MODIFY 1 trp [search]]
    [POP11
        prb_run(TRP_search, sim_myself.sim_data, false);
    ]
    [STOP]

;;; last rule indicates failure

rule TR_grab_wall_last
    [MOTIVE motive == status active == trp [stop] ==]
     ==>
    [STOP]
enddefine;

;;; TRP_make_wall ------------------------------------------

;;; utilities ---------------------------------------------------------

lvars X1, Y1, X2, Y2, H1, H2;
define proc_m2();
;;;    not(X1 = X2) or not(Y1 = Y2) and
    not(close_enough(X1, Y1, X2, Y2)) and
        not(is_right_orientation(H1, H2))
enddefine;

;;; rules -------------------------------------------------------------

define :ruleset TRP_make_wall;

rule TR_make_wall_r1
    [MOTIVE motive == status active == trp [make_wall ?X ?Y ?Heading ?Side] ==]
    [belief name ?Side wall is ?Obj [at ?XX ?YY ?Heading] ==]
    [WHERE close_enough(X, Y, XX, YY) ]
     ==>
    [MODIFY 1 trp [done make_wall ?X ?Y ?Heading ?Side]]
    [STOP]

rule TR_make_wall_r2
    [MOTIVE motive == status active == trp [make_wall ?X ?Y ?Heading ?Side] ==]
    [new_sense_data == name ?Obj type bar status alive == x ?XX y ?YY
        == held false == heading ?Hheading ==]
    [WHERE close_enough(X, Y, XX, YY) and
        is_right_orientation(Heading, Hheading) ]
     ==>
    [NOT belief name ?Side ==]
    [belief name ?Side wall is ?Obj [at ?X ?Y ?Heading]
        [m_conditions [
            [belief name = wall is ?Obj [at ?X1 ?Y1 ?H1] ==]
            [belief == name ?Obj == x ?X2 y ?Y2 == heading ?H2 ==]
            [WHERE proc_m2() ] ] ] ]
    [MODIFY 1 trp [stop]]
    [STOP]

rule TR_make_wall_r3
    [MOTIVE motive == status active == trp [make_wall ?X ?Y ?Heading ?Side] ==]
    [held ?Obj]
    [NOT belief name = wall is ?Obj ==]
    [belief == name ?Obj type bar status alive == held true ==]
     ==>
    [MODIFY 1 trp [place_bar ?Obj ?X ?Y ?Heading]]
    [POP11
        prb_run(TRP_place_bar, sim_myself.sim_data, false);
    ]
    [STOP]

rule TR_make_wall_r4
    [MOTIVE motive == status active == trp [make_wall ?X ?Y ?Heading ?Side] ==]
     ==>
    [MODIFY 1 trp [grab_wall]]
    [POP11
        prb_run(TRP_grab_wall, sim_myself.sim_data, false);
    ]
    [STOP]

rule TR_make_wall_last
    [MOTIVE motive == status active == trp [stop] ==]
     ==>
    [STOP]

enddefine;

;;; TRP_shut_door -----------------------------------------------

;;; utilities ---------------------------------------------------------

;;; rules -------------------------------------------------------------

define :ruleset TRP_shut_door;

rule TR_shut_door_r1
    [MOTIVE motive == status active == trp [shut_door] ==]
    [belief name north wall is ==]
    [belief name east wall is ==]
    [belief name south wall is ==]
    [belief name west wall is ?Obj [at ?X ?Y ?Heading] ==]
    [WHERE
        close_enough(X, Y,
            front(ap_room_loc) - ap_offset, back(ap_room_loc)) and
        is_right_orientation(90, Heading) ]
    [NOT held ?Obj]
     ==>
    ;;; agent believes all walls of room are in place
    [MODIFY 1 trp [done shut_door]]
    [STOP]

rule TR_shut_door_r2
    [MOTIVE motive == status active == trp [shut_door] ==]
    [belief name north wall is ==]
    [belief name east wall is ==]
    [belief name south wall is ==]
    [belief name west wall is ?Obj]
    [new_sense_data == name ?Obj type bar status alive == x ?X y ?Y
        == held false == heading ?Heading ==]
    [WHERE
        close_enough(X, Y,
            front(ap_room_loc) - ap_offset, back(ap_room_loc)) and
        is_right_orientation(90, Heading) ]
    [NOT held ?Obj]
     ==>
    [NOT belief name west wall is ==]
    [belief name west wall is ?Obj [at ?X ?Y ?Heading]
        [m_conditions [
            [belief name = wall is ?Obj [at ?X1 ?Y1 ?H1] ==]
            [belief == name ?Obj == x ?X2 y ?Y2 == heading ?H2 ==]
            [WHERE proc_m2() ] ] ] ]
    [MODIFY 1 trp [stop]]
    [STOP]

rule TR_shut_door_r3
    [MOTIVE motive == status active == trp [shut_door] ==]
    [belief name north wall is ==]
    [belief name east wall is ==]
    [belief name south wall is ==]
    [belief name west wall is ?Obj ==]
    [held ?Obj]
     ==>
    [LVARS x1 y1 heading side]
    [POP11
        front(ap_room_loc) - ap_offset -> x1;
        back(ap_room_loc) -> y1;
        90 -> heading;
    ]
    [NOT belief name west wall is ?Obj ==]
    [belief name west wall is ?Obj]
        ;;; required to prevent belief maintenance removing this belief
    [MODIFY 1 trp [place_bar ?Obj ?x1 ?y1 ?heading]]
    [POP11
        prb_run(TRP_place_bar, sim_myself.sim_data, false);
    ]
    [STOP]

rule TR_shut_door_r4
    [MOTIVE motive == status active == trp [shut_door] ==]
    [belief name north wall is ==]
    [belief name east wall is ==]
    [belief name south wall is ==]
    [belief name west wall is ?Obj ==]
     ==>
    [MODIFY 1 trp [grab_object ?Obj]]
    [POP11
        prb_run(TRP_grab_object, sim_myself.sim_data, false);
    ]
    [STOP]

rule TR_shut_door_last
    [MOTIVE motive == status active == trp [stop] ==]
     ==>
    [STOP]
enddefine;

;;; TRP_open_door -----------------------------------------------

;;; utilities ---------------------------------------------------------

;;; rules -------------------------------------------------------------

define :ruleset TRP_open_door;

rule TR_open_door_r1
    [MOTIVE motive == status active == trp [open_door] ==]
    [belief name north wall is ==]
    [belief name east wall is ==]
    [belief name south wall is ==]
    [belief name west wall is ?Obj [at ?X ?Y ?Heading] ==]
    [WHERE
        close_enough(X, Y,
            front(ap_room_loc) - ap_offset - 4, back(ap_room_loc)) and
        is_right_orientation(65, Heading) ]
    [NOT held ?Obj]
     ==>
    [MODIFY 1 trp [done open_door]]
    [STOP]

rule TR_open_door_r2
    [MOTIVE motive == status active == trp [open_door] ==]
    [belief name north wall is ==]
    [belief name east wall is ==]
    [belief name south wall is ==]
    [belief name west wall is ?Obj]
    [new_sense_data == name ?Obj type bar status alive == x ?X y ?Y
        == held false == heading ?Heading ==]
    [WHERE
        close_enough(X, Y,
            front(ap_room_loc) - ap_offset - 4, back(ap_room_loc)) and
        is_right_orientation(65, Heading) ]
    [NOT held ?Obj]
     ==>
    [NOT belief name west wall is ==]
    [belief name west wall is ?Obj [at ?X ?Y ?Heading]
        [m_conditions [
            [belief name = wall is ?Obj [at ?X1 ?Y1 ?H1] ==]
            [belief == name ?Obj == x ?X2 y ?Y2 == heading ?H2 ==]
            [WHERE proc_m2() ] ] ] ]
    [MODIFY 1 trp [stop]]
    [STOP]

rule TR_open_door_r3
    [MOTIVE motive == status active == trp [open_door] ==]
    [belief name north wall is ==]
    [belief name east wall is ==]
    [belief name south wall is ==]
    [belief name west wall is ?Obj ==]
    [held ?Obj]
     ==>
    [LVARS x1 y1 heading side]
    [POP11
        front(ap_room_loc) - ap_offset - 4 -> x1;
        back(ap_room_loc) -> y1;
        65 -> heading;
    ]
    [NOT belief name west wall is ?Obj ==]
    [belief name west wall is ?Obj]
        ;;; required to prevent belief maintenance removing this belief
    [MODIFY 1 trp [place_bar ?Obj ?x1 ?y1 ?heading]]
    [POP11
        prb_run(TRP_place_bar, sim_myself.sim_data, false);
    ]
    [STOP]

rule TR_open_door_r4
    [MOTIVE motive == status active == trp [open_door] ==]
    [belief name north wall is ==]
    [belief name east wall is ==]
    [belief name south wall is ==]
    [belief name west wall is ?Obj ==]
     ==>
    [MODIFY 1 trp [grab_object ?Obj]]
    [POP11
        prb_run(TRP_grab_object, sim_myself.sim_data, false);
    ]
    [STOP]

rule TR_open_door_last
    [MOTIVE motive == status active == trp [stop] ==]
     ==>
    [STOP]
enddefine;

;;; TRP_charge_object -----------------------------------------------

;;; utilities ---------------------------------------------------------

;;; rules -------------------------------------------------------------

define :ruleset TRP_charge_object;

rule TR_charge_object_r1
    [MOTIVE motive == status active == trp [charge_object ?Obj ?Level] ==]
    [NOT held ?Obj]
    [new_sense_data == name ?Obj == charge ?Level2 ==]
    [WHERE Level2 >= Level ]
     ==>
    [MODIFY 1 trp [done charge_object ?Obj ?Level]]
    [STOP]

rule TR_charge_object_r2
    [MOTIVE motive == status active == trp [charge_object ?Obj ?Level] ==]
    [held ?Obj]
    [new_sense_data == name ?Obj == charge ?Level2 ==]
    [WHERE Level2 >= Level ]
     ==>
    [MODIFY 1 trp [drop ?Obj]]
    [POP11
        prb_run(TRP_drop, sim_myself.sim_data, false);
    ]
    [STOP]

rule TR_charge_object_r3
    [MOTIVE motive == status active == trp [charge_object ?Obj ?Level] ==]
    [held ?Obj]
    [new_sense_data == name ?Obj == x ?X y ?Y ==]
    [new_sense_data == name gas1 == x ?XX y ?YY ==]
    [WHERE close_enough(X, Y, XX, YY) ]
     ==>
    [closure charge ?Obj gas1]
    [MODIFY 1 trp [stop]]
    [STOP]

rule TR_charge_object_r4
    [MOTIVE motive == status active == trp [charge_object ?Obj ?Level] ==]
     ==>
    [MODIFY 1 trp [take_object ?Obj gas1]]
    [POP11
        prb_run(TRP_take_object, sim_myself.sim_data, false);
    ]
    [STOP]

rule TR_charge_object_last
    [MOTIVE motive == status active == trp [stop] ==]
     ==>
    [STOP]
enddefine;

;;; TRP_dismiss_object -----------------------------------------------

;;; utilities ---------------------------------------------------------

;;; rules -------------------------------------------------------------

define :ruleset TRP_dismiss_object;

rule TR_dismiss_object_r1
    [MOTIVE motive == status active == trp [dismiss_object ?Obj] ==]
    [not_exists ?Obj]
     ==>
    [MODIFY 1 trp [done dismiss_object ?Obj]]
    [STOP]

rule TR_dismiss_object_r2
    [MOTIVE motive == status active == trp [dismiss_object ?Obj] ==]
    [held ?Obj]
    [new_sense_data == name ?Obj == x ?X y ?Y ==]
    [new_sense_data == name exit1 == x ?XX y ?YY ==]
    [WHERE close_enough(X, Y, XX, YY) ]
     ==>
    [closure dismiss ?Obj exit1]
    [MODIFY 1 trp [stop]]
    [STOP]

rule TR_dismiss_object_r3
    [MOTIVE motive == status active == trp [dismiss_object ?Obj] ==]
     ==>
    [MODIFY 1 trp [take_object ?Obj exit1]]
    [POP11
        prb_run(TRP_take_object, sim_myself.sim_data, false);
    ]
    [STOP]

rule TR_dismiss_object_last
    [MOTIVE motive == status active == trp [stop] ==]
     ==>
    [STOP]
enddefine;

;;; TRP_put_safe-------------------------------------------------------

;;; utilities ---------------------------------------------------------

;;; rules -------------------------------------------------------------

define :ruleset TRP_put_safe;

rule TR_put_safe_r1
    [MOTIVE motive == status active == trp [put_safe ?Obj1 ?Obj2] ==]
    [belief == name ?Obj1 == x ?X1 y ?Y1 == poly_space ?Polygon ==]
    [belief == name ?Obj2 == x ?X2 y ?Y2 ==]
        ;;; object to move away from is a bar, ditch or wall
    [WHERE not(is_near_ditch(Polygon, X1, Y1, X2, Y2, near_ditch_proximity + 5)) ]
    [NOT held ?Obj2]
==>
    [MODIFY 1 trp [done put_safe ?Obj1 ?Obj2]]
    [STOP]

rule TR_put_safe_r2
    [MOTIVE motive == status active == trp [put_safe ?Obj1 ?Obj2] ==]
    [belief == name ?Obj1 == x ?X1 y ?Y1 ??End]
    [WHERE not(End matches [== poly_space ==]) ]
    [belief == name ?Obj2 == x ?X2 y ?Y2 ==]
    [WHERE
        get_the_distance(conspair(X1,Y1), conspair(X2,Y2))
            > near_ditch_proximity + 5 ]
    [NOT held ?Obj2]
==>
    [MODIFY 1 trp [done put_safe ?Obj1 ?Obj2]]
    [STOP]

rule TR_put_safe_r3
    [MOTIVE motive == status active == trp [put_safe ?Obj1 ?Obj2] ==]
    [belief == name ?Obj1 == x ?X1 y ?Y1 == poly_space ?Polygon ==]
    [belief == name ?Obj2 == x ?X2 y ?Y2 ==]
        ;;; object to move away from is a bar, ditch or wall
    [held ?Obj2]
    [WHERE not(is_near_ditch(Polygon, X1, Y1, X2, Y2, near_ditch_proximity + 5)) ]
==>
    [MODIFY 1 trp [drop ?Obj2]]
    [POP11
        prb_run(TRP_drop, sim_myself.sim_data, false);
    ]
    [STOP]

rule TR_put_safe_r4
    [MOTIVE motive == status active == trp [put_safe ?Obj1 ?Obj2] ==]
    [belief == name ?Obj1 == x ?X1 y ?Y1 ??End]
    [WHERE not(End matches [== poly_space ==]) ]
    [belief == name ?Obj2 == x ?X2 y ?Y2 ==]
    [held ?Obj2]
    [WHERE
        get_the_distance(conspair(X1,Y1), conspair(X2,Y2))
            > near_ditch_proximity + 5 ]
==>
    [MODIFY 1 trp [drop ?Obj2]]
    [POP11
        prb_run(TRP_drop, sim_myself.sim_data, false);
    ]
    [STOP]

rule TR_put_safe_r5
    [MOTIVE motive == status active == trp [put_safe ?Obj1 ?Obj2] ==]
    [belief == name ?Obj1 == x ?X1 y ?Y1 ==]
    [belief == name ?Obj2 == x ?X2 y ?Y2 ==]
    [held ?Obj2]
==>
    [LVARS newx newy]
    [POP11 world_size div 2 ->> newx -> newy; ]
    [MODIFY 1 trp [amble ?newx ?newy]]
    [POP11
        prb_run(TRP_amble, sim_myself.sim_data, false);
    ]
    [STOP]

rule TR_put_safe_r6
    [MOTIVE motive == status active == trp [put_safe ?Obj1 ?Obj2] ==]
    [belief == name ?Obj1 == x ?X1 y ?Y1 ==]
==>
    [MODIFY 1 trp [grab_object ?Obj2]]
    [POP11
        prb_run(TRP_grab_object, sim_myself.sim_data, false);
    ]
    [STOP]

rule TR_put_safe_r7
    [MOTIVE motive == status active == trp [put_safe ?Obj1 ?Obj2] ==]
==>
    ;;; something wrong as agent hasn't the necessary info
    ;;; to complete the plan
    [SAY 'Error in TRP_put_safe']
    [MODIFY 1 trp [done put_safe ?Obj1 ?Obj2]]
    [STOP]

rule TR_put_safe_last
    ==>
    [STOP]
enddefine;

;;; trsim_action_rules--------------------------------------------

;;; utilities-----------------------------------------------------

define :method trsim_closure_set_heading( a:trsim_actor_agent, angle);
lvars a, angle;

    angle -> a.trsim_heading;

enddefine;

define :method trsim_closure_setspeed( a:trsim_actor_agent, spd);
lvars a, spd;

    spd -> a.trsim_speed;

enddefine;

define :method trsim_closure_charge( c:trsim_actor_agent, a:trsim_agent,
        b:trsim_agent );
lvars a, b, c;

    if b.sim_name == "gas1" then
        ;;; ok
        if close_enough(
            destpair(a.trsim_location), destpair(b.trsim_location)) then
            ;;; ok
            if a.trsim_held then
                ;;; ok, do it
                a.trsim_charge + 100 -> a.trsim_charge
            endif;
        endif;
    endif;

enddefine;

define :method trsim_closure_dismiss( a:trsim_actor_agent,
    b:trsim_agent, c:trsim_agent );
lvars a, b, c, agn;

    if c.sim_name == "exit1" then
        if close_enough(
            destpair(b.trsim_location), destpair(c.trsim_location)) then
            if b.trsim_held then
                ;;; dismiss object a from domain
                [^b] -> sim_object_delete_list;
                sim_add_data([not_exists %b.sim_name%], a.sim_data);
                    ;;; if agent dismisses something, it automatically
                    ;;; knows it no longer exists
                sim_delete_data([held %b.sim_name%], a.sim_data);
                    ;;; now also remove agent from graphics list
                conspair(-50,-50) -> b.trsim_location;
                draw_move_agent(b);
            endif;
        endif;
    endif;
enddefine;

define :method trsim_closure_rotate_bar( a:trsim_actor_agent,
        thebar:trsim_bar_agent, amnt );
lvars a, thebar, amnt;
lconstant max_rot = 10;

    if amnt >= max_rot then max_rot -> amnt endif;
    if amnt < 0 then
        [Cannot rotate a -ve angle] =>
        interrupt();
    endif;
    unless not(thebar.trsim_held) then
        rotate_angle(thebar.trsim_heading, amnt) -> thebar.trsim_heading;
        draw_move_agent(thebar);
    else
        [Attempting to rotate an object that is not held] =>
        interrupt();
    endunless;

enddefine;

define :method trsim_closure_drop( a:trsim_actor_agent, obj:trsim_agent );
lvars a, obj, newx, newy;

    unless not(obj.trsim_held) then
        false -> obj.trsim_held;
            ;;; oops! a dead bug will be reborn by being dropped!
        sim_delete_data([held %obj.sim_name%], a.sim_data);
            ;;; will need changing if holds more than one item
        return_new_point(
            destpair(a.trsim_location), a.trsim_heading, 0.2) -> (newx,newy);
        conspair(newx, newy) -> obj.trsim_location;

;;;        copy(a.trsim_location) -> obj.trsim_location;
        draw_move_agent(obj);
    else
        [Attempting to drop an object that is not held] =>
        interrupt();
    endunless;

enddefine;

define :method trsim_closure_grab_object( a:trsim_actor_agent, obj:trsim_agent );
lvars a, obj;

    unless obj.trsim_held then
        unless
            not(close_enough(
                destpair(a.trsim_location), destpair(obj.trsim_location)))
            then
            true -> obj.trsim_held;
            a.trsim_location -> obj.trsim_location;
            draw_move_agent(obj);
            sim_add_data([held ^(obj.sim_name)], a.sim_data);
                ;;; will need changing if holds more than one item
        else
            ;;; [NOT CLOSE ENOUGH TO OBJECT] =>
        endunless;
    endunless;
enddefine;

;;; if result is false, the other values are ignored
define procedure is_legal_move( ag_myself, x1, y1, x2, y2 )
                -> (result, dist, newx, newy) ;
    lvars newpoint, newx, newy, result, ag, ag_myself, temp_list, dist, dist1;

    [%
        ;;; all_agents may differ from sim_objects
        for ag in all_agents do
            unless ag == ag_myself then
                if istrsim_bar_agent(ag) then
                    unless ag.trsim_held then
                        ;;; check if hit
                        collision_detect(x1, y1, x2, y2, ag.bar_current_poly)
                            -> result -> newpoint;
                        if result then newpoint endif;
                    endunless;
                endif;
            endunless;
        endfor;
    %] -> temp_list;
    if null(temp_list) then
        false -> result;
        x2, y2 -> (newx, newy);
    else
        true -> result;
        false ->> (newx, newy);
        lvars coords;
        for coords in temp_list do
            if not(newx) then
                ;;; Something wrong here. Computing distance of
                ;;; point from itself???

                explode(coords) -> (newx, newy);
                sim_distance_from(explode(coords), newx, newy) -> dist1;
            else
                explode(coords) -> (newx, newy);
                sim_distance_from(x1, y1, newx, newy) -> dist;
                if dist < dist1 then
                    destpair(coords) -> (newx, newy);
                    dist -> dist1;
                endif;
            endif;
        endfor;
        dist1 -> dist;
    endif;

enddefine;

define :method trsim_closure_move( a:trsim_agent, holding_list );
    lvars a, x1, y1, x2, y2, newx, newy, result, dist,
        anobj, temp, holding_list, item;

    unless a.trsim_held or (a.trsim_charge < 0) then
        destpair(a.trsim_location) -> (x1, y1);
        x1 + a.trsim_speed * cos(a.trsim_heading) -> x2;
        y1 + a.trsim_speed * sin(a.trsim_heading) -> y2;
        ;;; drain charge
        if istrsim_bug_agent(a) then
            a.trsim_charge - 1 -> a.trsim_charge
        endif;

        ;;; now check for collision detection
        is_legal_move(a, x1, y1, x2, y2) -> (result, dist, newx, newy);
        unless result then
            conspair(x2,y2) -> a.trsim_location;
            draw_move_agent(a);
        else
            ;;; a collision has occurred
            dist / bounce_ratio -> temp;
            dist - temp -> dist;
            return_new_point(
                destpair(a.trsim_location), a.trsim_heading, dist)
                    -> (newx, newy);
            unless dist >= 0 and istrsim_actor_agent(a) then
            ;;; Why is nothing done with the output of is_legal_move ????
                is_legal_move(a, x1, y1, newx, newy)
                    -> (result, temp, newx, newy) ;
/*
                if temp then
                    [Second collision] =>
                    interrupt();
                endif;
*/
            else
                ;;; dist is 0
                newx + random(0.1) - random(0.1) -> newx;
                newy + random(0.1) - random(0.1) -> newy;
            endunless;
            conspair(newx,newy) -> a.trsim_location;
            random(360) -> a.trsim_heading;
            draw_move_agent(a);
        endunless;
        ;;; also move objects that are currently held
        unless null(holding_list) then
            copy(a.trsim_location) -> trsim_location(valof(front(holding_list)));
            draw_move_agent(valof(front(holding_list)));
        endunless;
    endunless;
enddefine;

;;; trsim_actor_action_rules------------------------------------------------

define :ruleset trsim_actor_action_rules;

rule trsim_do_move_closure_a
    [closure move]
    [NOT held ==]
    ==>
    [do trsim_closure_move []]
    [NOT closure move]

rule trsim_do_move_closure_holding
    [closure move]
    [held ??Objs]
    ==>
    [do trsim_closure_move ?Objs]
    [NOT closure move]

rule trsim_do_rotate_closure_a
    [closure rotate ?To]
    ==>
    [do trsim_closure_set_heading ?To]
    [NOT closure rotate ?To]

rule trsim_do_setspeed
    [closure setspeed ?To]
    ==>
    [do trsim_closure_setspeed ?To]
    [NOT closure setspeed ?To]

rule trsim_do_grab_object_closure
    [closure grab_object ?It]
    ==>
    [LVARS obj]
    [POP11 valof(It) -> obj; ]
    [do trsim_closure_grab_object ?obj]
    [NOT closure grab_object ?It]

rule trsim_do_drop_closure
    [closure drop ?It]
    ==>
    [LVARS obj]
    [POP11 valof(It) -> obj; ]
    [do trsim_closure_drop ?obj]
    [NOT closure drop ?It]

rule trsim_do_rotate_bar_closure
    [closure rotate_bar ?It ?To]
    ==>
    [LVARS obj]
    [POP11 valof(It) -> obj; ]
    [do trsim_closure_rotate_bar ?obj ?To]
    [NOT closure rotate_bar ?It ?To]

rule trsim_do_charge_closure
    [closure charge ?It ?With]
    ==>
    [LVARS obj obj2]
    [POP11 valof(It) -> obj; valof(With) -> obj2; ]
    [do trsim_closure_charge ?obj ?obj2]
    [NOT closure charge ?It ?With]

rule trsim_do_dismiss_closure
    [closure dismiss ?It ?With]
    ==>
    [LVARS [obj = valof(It)][obj2 = valof(With)]]
    [do trsim_closure_dismiss  ?obj ?obj2]
    [NOT closure dismiss ?It ?With]

rule act_last_a
     ==>
    [POP11 sim_myself.trsim_clock + 1 -> sim_myself.trsim_clock;]
    [STOP]
enddefine;
