;;; trsim_bug_rules-------------------------------------------------

;;; utilities-------------------------------------------------------

define procedure isa_set(setlist) -> bool;
lvars bool = true, item, newlist;

    for item in setlist do
        delete(item, setlist) -> newlist;
        if length(newlist) - length(setlist) < -1 then
            false -> bool;
            quitloop;
        endif;
    endfor;

enddefine;

/*
    isa_set([a b c d e]) =>
    isa_set([a b c d c c]) =>
*/

;;; rules-----------------------------------------------------------

define :ruleset trsim_bug_rules;

;;; if the bug is surrounded by three or more other bugs
;;; it turns into a thug

rule attack_bug
    [NOT closure ==]
    [state thug ==]
    [new_sense_data == name ?Obj type bug status ?Status == x ?X y ?Y ==]
    [WHERE (Status /= "thug") and (Status /= "dead") ]
    [WHERE Obj /= sim_myself.sim_name ]
==>
    ;;; change heading and move to attack bug
    [LVARS ch_heading ]
    [POP11
        lvars (xself, yself) = destpair(sim_myself.trsim_location);
        course(xself, yself, X, Y) -> ch_heading;
        if random(10) <= 1 then
            sim_add_data([closure move], sim_myself.sim_data)
        endif;
    ]
    [closure rotate set ?ch_heading]

rule search_attack
    [NOT closure]
    [state thug ==]
    [NOT_EXISTS
        [new_sense_data == name ?Obj type bug status ?Status == x ?X y ?Y ==]
        [WHERE (Status /= "thug") and (Status /= "dead") ]
        [WHERE Obj /= sim_myself.sim_name ]
    ]
==>
    [LVARS ch_heading]
    [POP11
        random(360) -> ch_heading;
        sim_add_data([closure move], sim_myself.sim_data)
    ]
    [closure rotate set ?ch_heading]

rule become_normal_attack
    [state thug ?Level]
    [WHERE (Level <= 0) and (sim_myself.trsim_status /= "dead") ]
    ==>
    [POP11 "alive" -> sim_myself.trsim_status;
            draw_move_agent(sim_myself); ]
    [NOT state thug ==]

rule become_normal_frightened
    [state frightened ?Level]
    [WHERE (Level <= 0) and (sim_myself.trsim_status /= "dead") ]
    ==>
    [POP11 "alive" -> sim_myself.trsim_status;
            draw_move_agent(sim_myself); ]
    [NOT state frightened ==]

rule become_thug
    [NOT state thug ==]
    [new_sense_data == name ?Obj1 type bug status alive ==]
        [WHERE Obj1 /= sim_myself.sim_name ]
    [new_sense_data == name ?Obj2 type bug status alive ==]
        [WHERE Obj2 /= sim_myself.sim_name ]
    [new_sense_data == name ?Obj3 type bug status alive ==]
        [WHERE Obj3 /= sim_myself.sim_name ]
;;;    [new_sense_data == name ?Obj4 type bug status alive ==]
;;;        [WHERE Obj4 /= sim_myself.sim_name ]
    [WHERE
        lvars objlist = [^Obj1 ^Obj2 ^Obj3
                        ;;; ^Obj4
                        ];
        isa_set(objlist)
    ]
    ==>
    [NOT state thug ==]
    [state thug 500] ;;; is number of cycles will be a thug
    [POP11 "thug" -> sim_myself.trsim_status;
           500 -> sim_myself.trsim_charge; ;;; becomes a superbug
            draw_move_agent(sim_myself);
    ]
    [STOP]

vars Distance;
rule become_frightened
    [NOT state thug ==]
    [new_sense_data == type bug status thug distance ?Distance ==]
    [WHERE Distance < 3 ]
==>
    ;;; become frightened
    [NOT state frightened ==]
    [state frightened 500]
    [POP11 "frightened" -> sim_myself.trsim_status;
        draw_move_agent(sim_myself);
    ]

rule calm_down_attack
    [state thug ?Level]
    [WHERE sim_myself.trsim_status /= "dead" ]
==>
    [POP11 Level - 1 -> Level;]
    [MODIFY 1 thug ?Level]

rule calm_down_fear
    [state frightened ?Level]
    [WHERE sim_myself.trsim_status /= "dead" ]
==>
    [POP11 Level - 1 -> Level;]
    [MODIFY 1 frightened ?Level]

;;; rules for determining thug behaviour

;;; rules for normal bug behaviour

rule bug_wander
    ;;; bug will default to random wander
    ;;; Insert conditions: bug alive and bug not held
    [NOT closure ==]
    [NOT state thug ==]
==>
    [POP11
        lvars rdm = random(10);
        if rdm <= 5 then
            sim_add_data([closure move], sim_myself.sim_data)
        endif;
    ]

rule bug_dither
    ;;; make bugs randomly change their heading
    ;;; Insert conditions: bug alive and bug not held
    [NOT state thug ==]
==>
    [LVARS ch_heading]
    [POP11
        lvars rdm = random(10);
        if rdm == 1 then
            'clockwise' -> ch_heading
        elseif rdm == 2 then
            'anticlockwise' -> ch_heading
        else
            'nochange' -> ch_heading
        endif;
    ]
    [closure rotate ?ch_heading]

rule bug_rules_last
    ;;; no more goals to process
     ==>
    [STOP]
enddefine;

;;; trsim_action_rules--------------------------------------------

;;; utilities-----------------------------------------------------

define :method trsim_closure_rotate_bug( a:trsim_bug_agent, rot_info );
lvars a, rot_info;
lconstant angle_of_rotation = 45;

    if rot_info = 'clockwise' then
        rotate_angle(trsim_heading(a), -angle_of_rotation) -> a.trsim_heading;
    elseif rot_info = 'anticlockwise' then
        ;;; anticlockwise
        rotate_angle(trsim_heading(a), angle_of_rotation) -> a.trsim_heading
    endif;

enddefine;

define :method trsim_closure_set_rotate( a:trsim_bug_agent, setting );
lvars a, setting;

    setting -> a.trsim_heading

enddefine;

;;; trsim_bug_action_rules------------------------------------------------

define :ruleset trsim_bug_action_rules;

rule trsim_do_move_closure_b
    [closure move]
==>
    [NOT closure move]
    [do trsim_closure_move []]

vars Direction;
rule trsim_do_rotate_closure_b
    [closure rotate ?Direction]
==>
    [NOT closure rotate ?Direction]
    [do trsim_closure_rotate_bug ?Direction]

rule trsim_set_rotate_closure
    [closure rotate set ?Direction]
==>
    [NOT closure rotate set ?Direction]
    [do trsim_closure_set_rotate ?Direction]

rule act_last_b
==>
    [POP11 sim_myself.trsim_clock + 1 -> sim_myself.trsim_clock;]
    [STOP]
enddefine;
