;;; gi.p
;;; ipw
;;; Graphical interface

uses rc_graphic;
uses rc_context;
uses popxlib;
uses xt_widgetinfo;
uses xpt_coretypes;

constant window_font = 2;

define :method update_xor( a:trsim_agent );
lvars a;
    a.trsim_location;
    a.trsim_status;
    a.trsim_heading;
enddefine;

define :method update_xor( a:trsim_actor_agent );
lvars a;
    a.trsim_location;
    a.trsim_status;
    a.trsim_heading;
    a.trsim_active_motive;
enddefine;

define :method update_xor( a:trsim_static_agent );
enddefine;

define :method update_xor( a:trsim_bar_agent );
    a.trsim_location;
    a.trsim_status;
    a.trsim_heading;
enddefine;

define :method update_xor( a:trsim_ditch_agent );
enddefine;

define procedure coords_to_window(coords) -> window_coords;

lvars coords, window_coords;

   [0 0] -> window_coords;
   coords(1) - (world_size div 2) -> window_coords(1);
   coords(2) - (world_size div 2) -> window_coords(2);

enddefine;

define :method draw_polygon( polygon:poly );

lvars polygon, count, c1, c2;

   for count from 1 to length(polygon.poly_points) do
      unless count = length(polygon.poly_points) then
         [%(polygon.poly_points)(count)(1)%
          %(polygon.poly_points)(count)(2)%] -> c1;
         [%(polygon.poly_points)(count + 1)(1)%
          %(polygon.poly_points)(count + 1)(2)%] -> c2;
         coords_to_window(c1) -> c1;
         coords_to_window(c2) -> c2;
         rc_drawline(c1(1), c1(2), c2(1), c2(2));
      else
         [%(polygon.poly_points)(count)(1)%
          %(polygon.poly_points)(count)(2)%] -> c1;
         [%(polygon.poly_points)(1)(1)%
          %(polygon.poly_points)(1)(2)%] -> c2;
         coords_to_window(c1) -> c1;
         coords_to_window(c2) -> c2;
         rc_drawline(c1(1), c1(2), c2(1), c2(2));
      endunless;
   endfor;

enddefine;

define :method draw_agent( a:trsim_agent );
lvars a, win_coords, x1, y1, old_xor;
dlocal rc_linewidth;
lvars xor = GXequiv;
dlocal rc_linefunction;
lconstant a_radius = 1;

    xor -> rc_linefunction;
    destpair(trsim_location(a)) -> (x1, y1);
    [^x1 ^y1] -> win_coords;
    coords_to_window(win_coords) -> win_coords;
    rc_print_at(win_coords(1) + 1, win_coords(2) + 1, a.trsim_id);
    if istrsim_actor_agent(a) then
        3 -> rc_linewidth;
        rc_circle(win_coords(1), win_coords(2), a_radius);
        rc_print_at(-55, -55, a.trsim_active_motive >< '');
    else
        switchon a.trsim_status
            case = "dead" then
                rc_print_at(win_coords(1) - 1, win_coords(2) - 1, 'x')
            case = "thug" then
                rc_print_at(win_coords(1) - 1, win_coords(2) - 1, 't')
            case = "frightened" then
                rc_print_at(win_coords(1) - 1, win_coords(2) - 1, 'f')
            else
                rc_circle(win_coords(1), win_coords(2), a_radius)
        endswitchon;
    endif;
enddefine;

define :method draw_agent( a:trsim_actor_agent );
lvars a, style, x1, y1, win_coords, a_radius;
dlocal rc_linefunction;
lvars xor = GXequiv;

    xor -> rc_linefunction;
    rc_linestyle -> style;
    LineOnOffDash -> rc_linestyle;
    ;;;LineDoubleDash -> rc_linestyle;
    destpair(trsim_location(a)) -> (x1, y1);
    [^x1 ^y1] -> win_coords;
    coords_to_window(win_coords) -> win_coords;
    last(front(a.sim_sensors)) -> a_radius;
    rc_circle(win_coords(1), win_coords(2), a_radius);
    style -> rc_linestyle;
    call_next_method(a);

enddefine;

define :method draw_move_agent( a:trsim_agent );
lvars a, new_xor, old_xor;

    [% update_xor(a) %] -> new_xor;
    a.trsim_xor -> old_xor;
    unless null(old_xor) then
        old_xor(1) -> a.trsim_location;
        old_xor(2) -> a.trsim_status;
        old_xor(3) -> a.trsim_heading;
        if length(old_xor) > 3 then
            old_xor(4) -> a.trsim_active_motive
        endif;
        draw_agent(a);
    endunless;
    new_xor(1) -> a.trsim_location;
    new_xor(2) -> a.trsim_status;
    new_xor(3) -> a.trsim_heading;
    if length(new_xor) > 3 then
        new_xor(4) -> a.trsim_active_motive
    endif;
    new_xor -> a.trsim_xor;
    draw_agent(a);
enddefine;

define :method draw_move_agent( a:trsim_static_agent );
lvars a;
    unless a.trsim_xor = false then
        draw_agent(a);
        false -> a.trsim_xor;
    endunless;
enddefine;

define :method draw_move_agent( a:trsim_ditch_agent );
lvars a;
    unless a.trsim_xor = false then
        draw_agent(a);
        false -> a.trsim_xor;
    endunless;
enddefine;

define :method draw_agent( a:trsim_static_agent );
    lvars a, win_coords, x1, y1;
    dlocal rc_linewidth;
    lvars xor = GXequiv;
    dlocal rc_linefunction;
    lconstant a_radius = 1;

    xor -> rc_linefunction;
    destpair(trsim_location(a)) -> (x1, y1);
    [^x1 ^y1] -> win_coords;
    coords_to_window(win_coords) -> win_coords;
;;;    rc_print_at(win_coords(1), win_coords(2), a.trsim_id);
    rc_jumpto(win_coords(1), win_coords(2));
    3 -> rc_linewidth;
    if a.trsim_id = 'F' then
        ;;; refuel point
        rc_draw_rectangle(2, 2);
    elseif a.trsim_id = 'E' then
        ;;; exit point
        rc_draw_oblong(6, 6, 2);
    endif;
enddefine;

define procedure rotate_angle(angle, rotation) -> new_angle;
;;; Given an angle and rotation angle returns a new
;;; angle within 0 - 360 degrees
lvars angle, rotation, new_angle;

    angle + rotation -> new_angle;
    if new_angle > 360 then
        new_angle - 360 -> new_angle
    elseif new_angle < 0 then
        360 + new_angle -> new_angle;
    endif;
enddefine;

define :method make_bar_points( a:trsim_bar ) -> bar_points;
lvars a, bar_points, alpha, new_heading, x, y, len;
lvars x1, y1, x2, y2, x3, y3, x4, y4;

    arctan(a.bar_width / a.bar_length) -> alpha;
    sqrt( (a.bar_length / 2)**2 + (a.bar_width / 2)**2 ) -> len;
    rotate_angle(a.trsim_heading, alpha) -> new_heading;
    destpair(a.trsim_location) -> (x,y);
    x + len * cos(new_heading) -> x1;
    y + len * sin(new_heading) -> y1;
    rotate_angle(a.trsim_heading, -alpha) -> new_heading;
    x + len * cos(new_heading) -> x2;
    y + len * sin(new_heading) -> y2;
    rotate_angle(a.trsim_heading, 180 + alpha) -> new_heading;
    x + len * cos(new_heading) -> x3;
    y + len * sin(new_heading) -> y3;
    rotate_angle(a.trsim_heading, 180 - alpha) -> new_heading;
    x + len * cos(new_heading) -> x4;
    y + len * sin(new_heading) -> y4;
    [ [^x1 ^y1] [^x2 ^y2] [^x3 ^y3] [^x4 ^y4] ] -> bar_points;

enddefine;

/*
    vars b;
    define :instance b:trsim_bar_agent; enddefine;
    b.bar_length =>
    b.bar_width =>
    b.trsim_heading =>
    make_bar_points(b) =>
    draw_agent(b);

    b.bar_current_poly =>
*/

define :method draw_small_polygon( polygon:poly );
    lvars polygon, count, c1, c2;
    [%(polygon.poly_points)(1)(1)%
        %(polygon.poly_points)(1)(2)%] -> c1;
    [%(polygon.poly_points)(3)(1)%
        %(polygon.poly_points)(3)(2)%] -> c2;
    coords_to_window(c1) -> c1;
    coords_to_window(c2) -> c2;
    rc_drawline(c1(1), c1(2), c2(1), c2(2));
enddefine;

define :method draw_agent( a:trsim_bar_agent );
lvars a, bar_points, polygon, end1, end2, end3, end4, temp;
lvars xor = GXequiv;
dlocal rc_linefunction;

    rc_linewidth -> temp;
    xor -> rc_linefunction;
    make_bar_points(a) -> bar_points;
    make_polygon(bar_points) -> polygon;
    bar_points(1) -> end1;
    bar_points(2) -> end2;
    bar_points(3) -> end3;
    bar_points(4) -> end4;
    conspair(end1(1), end1(2)) -> a.bar_end1a_location;
    conspair(end2(1), end2(2)) -> a.bar_end1b_location;
    conspair(end3(1), end3(2)) -> a.bar_end2a_location;
    conspair(end4(1), end4(2)) -> a.bar_end2b_location;
    polygon -> a.bar_current_poly;
    if (a.trsim_id = 'bar') and (a.bar_width = 0) then
        4 -> rc_linewidth;
        draw_small_polygon(polygon);
    else
        draw_polygon(polygon)
    endif;
    temp -> rc_linewidth;

enddefine;

define :method draw_agent( a:trsim_ditch_agent );
lvars a, bar_points, polygon, end1, end2, style;
dlocal rc_linewidth;
lvars xor = GXequiv;
dlocal rc_linefunction;
lconstant lw = 2;

    rc_linestyle -> style;
    LineOnOffDash -> rc_linestyle;
    ;;;LineDoubleDash -> rc_linestyle;
    lw -> rc_linewidth;
    xor -> rc_linefunction;
    make_bar_points(a) -> bar_points;
    make_polygon(bar_points) -> polygon;
    bar_points(1) -> end1;
    bar_points(3) -> end2;
    polygon -> a.bar_current_poly;
    draw_polygon(polygon);
;;;    clearstack();
    style -> rc_linestyle;

enddefine;


define procedure kill;

   false -> rc_window;
   sysgarbage();

enddefine;

define procedure rc_set_background(colour);

lvars colour;

   colour -> rc_window("background");
   XpwClearWindow(rc_window);

enddefine;

define procedure rc_set_ink_color(color);

lvars color;

   XpwSetColor(rc_window, color) ->;

enddefine;

define procedure create_window;

   false -> rc_window;
   if window_size = 350 then
       rc_new_window(window_size, window_size, 680, 0, true)
   else
       rc_new_window(window_size, window_size, 350, 0, true)
   endif;
   ;;; set_window_font(window_font);
   rc_stretch_frame_by((window_size - (window_size * 0.1)) div world_size);
   ;;;rc_set_background('White');
   rc_set_background('white');
   ;;;rc_set_ink_color('grey0');
   rc_start();

enddefine;

define procedure set_window_font(font_number);

lvars font_number, font;

   lvars fonts =
      {'-adobe-helvetica-bold-r-normal-*-*-100-*-*-p-*-*-*'
      '-adobe-helvetica-bold-r-normal-*-*-120-*-*-p-*-*-*'
      '-adobe-helvetica-bold-r-normal-*-*-140-*-*-p-*-*-*'
      '8x13bold'
      '9x15bold'};
   fi_check(font_number, 1, datalength(fonts)) -> ;
   subscrv(font_number, fonts) -> font;
   unless XpwSetFont(rc_window, font) then
      mishap('BAD FONT', [^font])
   endunless;

enddefine;


define procedure window_to_coords(window_coords) -> coords;
lvars window_coords, coords, x, y;

    window_coords(1) + (world_size div 2) -> x;
    window_coords(2) + (world_size div 2) -> y;
    [^x ^y] -> coords;

enddefine;

/*
    coords_to_window([32 -33]) =>
    window_to_coords([-18 -83]) =>
*/
